"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import shutil

import numpy as np
from ase.build import bulk

from ... import context  # noqa
from mlacs.mlip import MliapDescriptor, LinearPotential


def test_parameters_snap():
    at = bulk("Si")

    rcut = 3.5
    parameters = dict(twojmax=5,
                      radelems=0.2,
                      welems=0.3)
    mliap = MliapDescriptor(at, rcut, parameters=parameters)

    assert mliap.ndesc == 20
    assert mliap.rcut == rcut
    assert mliap.chemflag == 0
    assert np.all(mliap.radelems == 0.2)
    assert np.all(mliap.welems == 0.3)

    at = bulk("FeRh", "cesiumchloride", 3.02)
    rcut = 3.5
    radelems = np.array([0.2, 0.3])
    welems = np.array([0.5, 0.3])
    parameters = dict(twojmax=5,
                      radelems=radelems,
                      welems=welems)
    mliap = MliapDescriptor(at, rcut, parameters=parameters)

    assert mliap.ndesc == 20
    assert mliap.rcut == rcut
    assert mliap.chemflag == 0
    assert np.all(mliap.radelems == radelems)
    assert np.all(mliap.welems == welems)

    at = bulk("FeRh", "cesiumchloride", 3.02)
    rcut = 3.5
    radelems = np.array([0.2, 0.3])
    welems = np.array([0.5, 0.3])
    parameters = dict(twojmax=5,
                      radelems=radelems,
                      welems=welems,
                      chemflag=1)
    mliap = MliapDescriptor(at, rcut, parameters=parameters)

    assert mliap.ndesc == 160
    assert mliap.rcut == rcut
    assert mliap.chemflag == 1
    assert np.all(mliap.radelems == radelems)
    assert np.all(mliap.welems == welems)

    at = bulk("FeRh", "cesiumchloride", 3.02)
    rcut = 3.5
    radelems = np.array([0.2, 0.3])
    welems = np.array([0.5, 0.3])
    parameters = dict(twojmax=5,
                      radelems=radelems,
                      welems=welems)
    mliap = MliapDescriptor(at, rcut, parameters=parameters, model="quadratic")

    assert mliap.ndesc == 230
    assert mliap.rcut == rcut
    assert np.all(mliap.radelems == radelems)
    assert np.all(mliap.welems == welems)


def test_parameters_so3():
    at = bulk("Si")

    rcut = 3.5
    radelems = np.array([0.2])
    welems = np.array([0.5])
    parameters = dict(nmax=2,
                      lmax=2,
                      radelems=radelems,
                      welems=welems)
    mliap = MliapDescriptor(at, rcut, parameters=parameters, style="so3")

    assert mliap.ndesc == 9
    assert mliap.rcut == rcut
    assert mliap.chemflag == 0
    assert np.all(mliap.radelems == radelems)
    assert np.all(mliap.welems == welems)

    at = bulk("FeRh", "cesiumchloride", 3.02)
    rcut = 3.5
    radelems = np.array([0.2, 0.3])
    welems = np.array([0.5, 0.3])
    parameters = dict(nmax=2,
                      lmax=2,
                      radelems=radelems,
                      welems=welems)
    mliap = MliapDescriptor(at, rcut, parameters=parameters, style="so3")

    assert mliap.ndesc == 9
    assert mliap.rcut == rcut
    assert mliap.chemflag == 0
    assert np.all(mliap.radelems == radelems)
    assert np.all(mliap.welems == welems)

    at = bulk("FeRh", "cesiumchloride", 3.02)
    rcut = 3.5
    radelems = np.array([0.2, 0.3])
    welems = np.array([0.5, 0.3])
    parameters = dict(nmax=2,
                      lmax=2,
                      radelems=radelems,
                      welems=welems)
    mliap = MliapDescriptor(at, rcut, parameters=parameters, model="quadratic",
                            style="so3")

    assert mliap.ndesc == 54
    assert mliap.rcut == rcut
    assert np.all(mliap.radelems == radelems)
    assert np.all(mliap.welems == welems)


model_ref = """# Si MLIP parameters
# Descriptor   snap

# nelems   ncoefs
1 21
   0.000000000000000000000000000000
   1.000000000000000000000000000000
   2.000000000000000000000000000000
   3.000000000000000000000000000000
   4.000000000000000000000000000000
   5.000000000000000000000000000000
   6.000000000000000000000000000000
   7.000000000000000000000000000000
   8.000000000000000000000000000000
   9.000000000000000000000000000000
  10.000000000000000000000000000000
  11.000000000000000000000000000000
  12.000000000000000000000000000000
  13.000000000000000000000000000000
  14.000000000000000000000000000000
  15.000000000000000000000000000000
  16.000000000000000000000000000000
  17.000000000000000000000000000000
  18.000000000000000000000000000000
  19.000000000000000000000000000000
  20.000000000000000000000000000000
"""


def test_writing_model():
    root = Path()
    folder = 'Mliap'
    mliapfold = root / folder

    at = bulk("Si")

    rcut = 3.5
    parameters = dict(twojmax=5)
    mliap = MliapDescriptor(at, rcut, parameters=parameters,
                            workdir=root, folder=folder)
    coeff = np.arange(0, 21)

    if mliapfold.exists():
        shutil.rmtree(mliapfold)
    mliapfold.mkdir()
    mliap.write_mlip(coeff)

    mliapfile = mliapfold / "MLIAP.model"
    assert mliapfile.exists()

    allref = model_ref.split("\n")
    allref = [line.rstrip() for line in allref]
    with open(mliapfile, "r") as fd:
        allpred = [line.rstrip() for line in fd]
    assert allref[0] == allpred[0]
    assert allref[1] == allpred[1]

    ref_info = [int(a) for a in allref[4].split()]
    pred_info = [int(a) for a in allpred[4].split()]
    assert np.all(ref_info == pred_info)

    for cref, cpred in zip(allref[5:], allpred[5:]):
        assert np.isclose(float(cref), float(cpred))

    shutil.rmtree(mliapfold)


desc_ref = """# Si MLIP parameters
# Descriptor:  snap
# Model:       linear

rcutfac         3.5
twojmax         5
rfac0           0.99363
rmin0           0.0
switchflag      1
bzeroflag       1
wselfallflag    0
bnormflag       1



nelems      1
elems       Si
radelems    0.5
welems      1.0
"""


def test_writing_descriptor():
    root = Path()
    folder = 'Mliap'
    mliapfold = root / folder

    at = bulk("Si")

    rcut = 3.5
    parameters = dict(twojmax=5)
    mliapfile = mliapfold / "MLIAP.descriptor"
    mliap = MliapDescriptor(at, rcut, parameters=parameters,
                            workdir=root, folder=folder)

    if mliapfold.exists():
        shutil.rmtree(mliapfold)
    mliapfold.mkdir()

    mliap._write_mlip_params()

    allref = desc_ref.split("\n")
    allref = [line.rstrip() for line in allref]

    with open(mliapfile, "r") as fd:
        allpred = [line.rstrip() for line in fd]

    assert allpred[0] == allref[0]
    assert allpred[1] == allref[1]
    assert allpred[2] == allref[2]

    for i in range(4, 12):
        ref = allref[i].split()
        pred = allref[i].split()

        assert ref[0] == pred[0]
        assert np.isclose(float(ref[1]), float(pred[1]), atol=1e-8)
    shutil.rmtree(mliapfold)


desc_ref_so3 = """# Si MLIP parameters
# Descriptor:  so3
# Model:       linear

rcutfac         3.5
nmax            3
lmax            2
alpha           1.0



nelems      1
elems       Si
radelems    0.5
welems      1.0
"""


def test_writing_descriptor_so3():
    root = Path()
    folder = 'Mliap'
    mliapfold = root / folder

    at = bulk("Si")

    rcut = 3.5
    parameters = dict(nmax=3, lmax=2)
    mliap = MliapDescriptor(at, rcut, parameters=parameters, style="so3",
                            workdir=root, folder=folder)
    mliapfile = mliapfold / "MLIAP.descriptor"

    if mliapfold.exists():
        shutil.rmtree(mliapfold)
    mliapfold.mkdir()

    mliap._write_mlip_params()

    allref = desc_ref_so3.split("\n")
    allref = [line.rstrip() for line in allref]

    with open(mliapfile, "r") as fd:
        allpred = [line.rstrip() for line in fd]

    assert allpred[0] == allref[0]
    assert allpred[1] == allref[1]
    assert allpred[2] == allref[2]

    for i in range(4, 8):
        ref = allref[i].split()
        pred = allpred[i].split()

        assert ref[0] == pred[0]
        assert np.isclose(float(ref[1]), float(pred[1]), atol=1e-8)

    for i in range(11, 15):
        ref = allref[i].split()
        pred = allpred[i].split()
        if ref[0] == "elems":
            assert ref[0] == pred[0]
            assert ref[1] == pred[1]
        else:
            assert ref[0] == pred[0]
            assert np.isclose(float(ref[1]), float(pred[1]), atol=1e-8)
    shutil.rmtree(mliapfold)


def test_get_pair_style_coeff():
    root = Path().absolute()
    at = bulk("Si")

    rcut = 3.5
    parameters = dict(twojmax=5)
    f = "Mliap"

    mliap = MliapDescriptor(at, rcut, parameters=parameters)
    mlip = LinearPotential(descriptor=mliap, folder=f)

    pred_st, pred_co = mlip.pair_style, mlip.pair_coeff

    model_file = (root / "Mliap/MLIAP.model").as_posix()
    desc_file = (root / "Mliap/MLIAP.descriptor").as_posix()
    ref_st = f"mliap model linear {model_file} descriptor sna {desc_file}"

    assert pred_st == ref_st
    ref_co = ["* * Si"]
    assert pred_co[0] == ref_co[0]

    # Now with so3
    f = "Mlip_So3"
    mliap = MliapDescriptor(at, rcut, parameters=parameters, style="so3")
    mlip = LinearPotential(descriptor=mliap, folder=f)

    pred_st, pred_co = mlip.pair_style, mlip.pair_coeff

    model_file = (root / "Mlip_So3/MLIAP.model").as_posix()
    desc_file = (root / "Mlip_So3/MLIAP.descriptor").as_posix()
    ref_st = f"mliap model linear {model_file} descriptor so3 {desc_file}"

    assert pred_st == ref_st
    ref_co = ["* * Si"]
    assert pred_co[0] == ref_co[0]



"""
// Copyright (C) 2022-2024 MLACS group (PR, AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ..core.manager import Manager
from ..utilities import get_elements_Z_and_masses

from ..state.lammps_state import (BaseLammpsState,
                                  LammpsState)


# ========================================================================== #
# ========================================================================== #
class ThermoState(BaseLammpsState):
    """
    Parent class for the thermodynamic state used in thermodynamic integration

    Parameters
    ----------
    atoms: :class:`ase.Atoms`
        ASE atoms object on which the simulation will be performed

    pair_style: :class:`str`
        pair_style for the LAMMPS input

    pair_coeff: :class:`str` or :class:`list` of :class:`str`
        pair_coeff for the LAMMPS input

    dt: :class:`int` (optional)
        Timestep for the simulations, in fs. Default ``1.5``

    nsteps: :class:`int` (optional)
        Number of production steps. Default ``10000``.

    nsteps_eq: :class:`int` (optional)
        Number of equilibration steps. Default ``5000``.

    nsteps_averaging: :class:`int` (optional)
        Number of step for equilibrate ideal structure
        at zero or finite pressure. Default ``10000``.

    rng: :class:`RNG object`
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    interval : :class:`int` (optional)
        Number of steps between log and traj writing. Override
        loginterval. Default ``50``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    workdir : :class:`str` (optional)
        Working directory for the LAMMPS MLMD simulations.
        If ``None``, a LammpsMLMD directory is created

    blocks : :class:`LammpsBlockInput` or :class:`list` (optional)
        Custom block input class. Can be a list of blocks.
        If ``None``, nothing is added in the input. Default ``None``.
    """
    def __init__(self,
                 atoms,
                 pair_style,
                 pair_coeff,
                 dt=1.5,
                 nsteps=10000,
                 nsteps_eq=5000,
                 nsteps_averaging=10000,
                 rng=None,
                 langevin=False,
                 logfile=None,
                 trajfile=None,
                 interval=500,
                 loginterval=50,
                 blocks=None,
                 neti=True,
                 **kwargs):

        super().__init__(nsteps,
                         nsteps_eq,
                         logfile,
                         trajfile,
                         loginterval=loginterval,
                         blocks=blocks,
                         neti=neti,
                         **kwargs)

        self.atoms = atoms
        self.pair_style = pair_style
        self.pair_coeff = pair_coeff
        self.nsteps = nsteps_averaging
        self.elem, self.Z, self.masses, self.charges = (
            get_elements_Z_and_masses(self.atoms))
        self.dt = dt

# ========================================================================== #
    @Manager.exec_from_path
    def run_averaging(self):
        """
        Get the right volume structure at finite pressure
        """
        eq_state = LammpsState(self.temperature,
                               self.pressure,
                               nsteps=self.nsteps,
                               nsteps_eq=self.nsteps_eq,
                               workdir=self.workdir,
                               folder=self.folder,
                               subfolder='Equilibration')

        self.atoms = eq_state.run_dynamics(self.atoms,
                                           self.pair_style,
                                           self.pair_coeff)

# ========================================================================== #
    def post_process(self):
        """
        """
        pass

# ========================================================================== #
    def _get_block_custom(self):
        """
        """
        return self._get_neti()



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
from pathlib import Path

import numpy as np
from ase.atoms import Atoms

from ..core.manager import Manager


# ========================================================================== #
# ========================================================================== #
class WeightingPolicy(Manager):
    """
    Parent class to manage weight in MLACS. This class define the standard to
    be used by the MLIP.

    Parameters
    ----------
    energy_coefficient: :class:`float`
        Weight of the energy in the fit
        Default 1.0

    forces_coefficient: :class:`float`
        Weight of the forces in the fit
        Default 1.0

    stress_coefficient: :class:`float`
        Weight of the stress in the fit
        Default 1.0

    database: :class:`ase.Trajectory`
        Initial database (optional)
        Default :class:`None`

    weight: :class:`list` or :class:`str`
        If you use an initial database, it needs weight.
        Can a list or an np.array of values or a file.
        Default :class:`None`

    """

    def __init__(self, energy_coefficient=1.0, forces_coefficient=1.0,
                 stress_coefficient=1.0, database=None, weight=None,
                 **kwargs):

        Manager.__init__(self, **kwargs)

        self.database = database
        self.matsize = []

        sum_efs = energy_coefficient + forces_coefficient + stress_coefficient
        self.energy_coefficient = energy_coefficient / sum_efs
        self.forces_coefficient = forces_coefficient / sum_efs
        self.stress_coefficient = stress_coefficient / sum_efs

        if database is not None:
            self.matsize = [len(a) for a in database]

        self.weight = np.array([])
        if weight is not None:
            if isinstance(weight, str):
                weight = np.loadtxt(weight)
            self.weight = weight
        elif (fname := self.subsubdir / 'MLIP.weight').exists():
            weight = np.loadtxt(fname)
            self.weight = weight
        else:
            self.weight = np.array([])

# ========================================================================== #
    def get_effective_conf(self):
        """
        Compute the number of effective configurations.
        """
        if len(self.weight) == 0:
            return 0
        neff = np.sum(self.weight)**2 / np.sum(self.weight**2)
        return neff

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef, f_mlipE):
        """
        """
        raise NotImplementedError

# ========================================================================== #
    def get_weights(self):
        """
        Return weighting matrices
        """
        w = self.init_weight()
        we, wf, ws = self.build_W_efs(w)
        we = we * self.energy_coefficient
        wf = wf * self.forces_coefficient
        ws = ws * self.stress_coefficient
        return np.r_[we, wf, ws]

# ========================================================================== #
    def init_weight(self, scale=1):
        """
        Scale the weight matrice to include the new configurations.
        Those have weight 1/Neff.
        """
        n_tot = len(self.matsize)
        neff = self.get_effective_conf()
        nnew = n_tot - len(self.weight)
        weight = (np.ones(n_tot)*scale) / (neff + nnew)
        ratio = neff / (neff + nnew)
        weight[:len(self.weight)] = self.weight * ratio
        return weight / np.sum(weight)

# ========================================================================== #
    def build_W_efs(self, w):
        """
        Transform W to W_efs.
        """
        w_e = w / np.sum(w)
        w_f = []
        w_s = []
        for i, n in enumerate(self.matsize):
            w_f.extend(w[i] * np.ones(3 * n) / (3 * n))
            w_s.extend(w[i] * np.ones(6) / 6)
        w_f = np.r_[w_f] / np.sum(np.r_[w_f])
        w_s = np.r_[w_s] / np.sum(np.r_[w_s])
        return w_e, w_f, w_s


# ========================================================================== #
# ========================================================================== #
class UniformWeight(WeightingPolicy):
    """
    Class that gives uniform weight in MLACS.

    Parameters
    ----------
    nthrow: :class:`int`
        Number of configurations to ignore when doing the fit.
        Three cases :

        1. If nconf > 2*nthrow, remove the nthrow first configuration
        2. If nthrow < nconf < 2*nthrow, remove the nconf-nthrow first conf
        3. If nconf < nthrow, keep all conf

    """

    def __init__(self, nthrow=0, energy_coefficient=1.0,
                 forces_coefficient=1.0, stress_coefficient=1.0,
                 database=None, weight=None, **kwargs):
        self.nthrow = nthrow
        WeightingPolicy.__init__(
                self,
                energy_coefficient=energy_coefficient,
                forces_coefficient=forces_coefficient,
                stress_coefficient=stress_coefficient,
                database=database, weight=weight, **kwargs)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef, f_mlipE):
        """
        Compute Uniform Weight taking into account nthrow :
        """
        fname = "MLIP.weight"
        if (filepath := Path(fname)).exists():
            filepath.unlink()

        nconf = len(self.matsize)
        to_remove = 0
        if nconf > 2*self.nthrow:
            to_remove = self.nthrow
        elif nconf > self.nthrow:
            to_remove = nconf-self.nthrow

        w = np.ones(nconf-to_remove) / (nconf-to_remove)
        w = np.r_[np.zeros(to_remove), w]
        self.weight = w

        header = "Using Uniform weighting\n"
        np.savetxt(fname, self.weight, header=header, fmt="%25.20f")
        return header, fname

# ========================================================================== #
    def update_database(self, atoms):
        """
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.matsize.extend([len(a) for a in atoms])


# ========================================================================== #
# ========================================================================== #
class IncreasingWeight(WeightingPolicy):
    """
    Class that gives increasing weight with the index of a configuration
    in MLACS.
    This weighting policy has been though for structural optimization.

    nthrow: :class: int
        Number of configurations to ignore when doing the fit.
        Three cases :
         1. If nconf > 2*nthrow, remove the nthrow first configuration
         2. If nthrow < nconf < 2*nthrow, remove the nconf-nthrow first conf
         3. If nconf < nthrow, keep all conf
    """

    def __init__(self, nthrow=0, power=1, energy_coefficient=1.0,
                 forces_coefficient=1.0, stress_coefficient=1.0,
                 database=None, weight=None):
        self.nthrow = nthrow
        self.power = power
        WeightingPolicy.__init__(
                self,
                energy_coefficient=energy_coefficient,
                forces_coefficient=forces_coefficient,
                stress_coefficient=stress_coefficient,
                database=database, weight=weight)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef, f_mlipE):
        """
        Compute Increasing Weight taking into account nthrow :
        """
        fname = "MLIP.weight"
        if (filepath := Path(fname)).exists():
            filepath.unlink()

        nconf = len(self.matsize)
        to_remove = 0
        if nconf > 2*self.nthrow:
            to_remove = self.nthrow
        elif nconf > self.nthrow:
            to_remove = nconf-self.nthrow

        w = (np.arange(nconf-to_remove, dtype=float) + 1)**self.power
        w /= w.sum()
        w = np.r_[np.zeros(to_remove), w]
        self.weight = w

        header = "Using Increasing weighting\n"
        np.savetxt(fname, self.weight, header=header, fmt="%25.20f")
        return header, fname

# ========================================================================== #
    def update_database(self, atoms):
        """
        Update the database.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.matsize.extend([len(a) for a in atoms])



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

import numpy as np
from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.mlip import SnapDescriptor, LinearPotential
from mlacs.state import LinearInterpolation, NebLammpsState, PafiLammpsState
from mlacs import OtfMlacs


@pytest.fixture
def expected_folder(expected_folder_base):
    return expected_folder_base


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


@pytest.mark.skipif(context.has_lammps_nompi(),
                    reason="Lammps needs mpi to run PAFI")
def test_mlacs_pafi_vanilla(root, treelink):

    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 5
    nconfs = 4
    nconfs_init = 1
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")

    nimages = 6
    neb = NebLammpsState(nebat, nimages=nimages)
    state = PafiLammpsState(300, neb, nsteps_eq=2, nsteps=100)

    sampling = OtfMlacs(nebat[0], state, calc, mlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    assert len(traj) == nstep
    # Check that the first atom is the one we started with
    assert traj[0] == nebat[0]
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms

    ml_energy = np.loadtxt(root / "MLIP-Energy_comparison.dat")
    ml_forces = np.loadtxt(root / "MLIP-Forces_comparison.dat")
    ml_stress = np.loadtxt(root / "MLIP-Stress_comparison.dat")

    nconfs = nconfs + nconfs_init
    assert ml_energy.shape == (nconfs, 2)
    assert ml_forces.shape == (nconfs * natoms * 3, 2)
    assert ml_stress.shape == (nconfs * 6, 2)

    # Check that spline is working well
    assert state.mep.patoms.splR.shape == (natoms, 3)
    assert state.mep.patoms.splDR.shape == (natoms, 3)
    assert state.mep.patoms.splD2R.shape == (natoms, 3)


@pytest.mark.slow
def test_mlacs_pafi_linear(root, treelink):

    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 5
    nconfs = 4
    nconfs_init = 1
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")

    nimages = 6
    # This is the setup to do BlueMoon Sampling
    neb = LinearInterpolation(nebat, xi=0.3, nimages=nimages, model=mlip)
    state = PafiLammpsState(300, neb, nsteps_eq=2, nsteps=100)

    sampling = OtfMlacs(nebat[0], state, calc, mlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    assert len(traj) == nstep
    # Check that the first atom is the one we started with

    for at in traj:
        assert len(at) == natoms

    ml_energy = np.loadtxt(root / "MLIP-Energy_comparison.dat")
    ml_forces = np.loadtxt(root / "MLIP-Forces_comparison.dat")
    ml_stress = np.loadtxt(root / "MLIP-Stress_comparison.dat")

    nconfs = nconfs + nconfs_init
    assert ml_energy.shape == (nconfs, 2)
    assert ml_forces.shape == (nconfs * natoms * 3, 2)
    assert ml_stress.shape == (nconfs * 6, 2)

    # Check that spline is working well
    assert state.mep.patoms.xi == 0.3
    assert state.mep.patoms.splR.shape == (natoms, 3)
    assert state.mep.patoms.splDR.shape == (natoms, 3)
    assert state.mep.patoms.splD2R.shape == (natoms, 3)
    assert np.all(state.mep.patoms.splD2R < 1e-8)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

import numpy as np
from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.mlip import SnapDescriptor, LinearPotential
from mlacs.state import LammpsState
from mlacs import OtfMlacs


@pytest.fixture
def expected_folder(expected_folder_base):
    return expected_folder_base


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


def test_mlacs_vanilla(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    natoms = len(atoms)
    nstep = 5
    nconfs = 4
    nconfs_init = 1
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")

    state = LammpsState(300, nsteps_eq=2, nsteps=100)

    sampling = OtfMlacs(atoms, state, calc, mlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")

    assert len(traj) == nstep
    # Check that the first atom is the one we started with
    assert traj[0] == atoms
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms

    ml_energy = np.loadtxt(root / "MLIP-Energy_comparison.dat")
    ml_forces = np.loadtxt(root / "MLIP-Forces_comparison.dat")
    ml_stress = np.loadtxt(root / "MLIP-Stress_comparison.dat")

    assert ml_energy.shape == (nconfs + nconfs_init, 2)
    assert ml_forces.shape == ((nconfs + nconfs_init) * natoms * 3, 2)
    assert ml_stress.shape == ((nconfs + nconfs_init) * 6, 2)


def test_mlacs_several_training(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    natoms = len(atoms)
    nsteps = 2
    nconfs = 1
    nconfs_init = 5
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")

    state = LammpsState(300, nsteps_eq=10, nsteps=100)

    sampling = OtfMlacs(atoms, state, calc, mlip, neq=5,
                        confs_init=nconfs_init)
    sampling.run(nsteps)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    assert len(traj) == nsteps
    # Check that the first atom is the one we started with
    assert traj[0] == atoms
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms

    traintraj = read(root / "Training_configurations.traj", ":")
    assert len(traintraj) == nconfs_init

    ml_energy = np.loadtxt(root / "MLIP-Energy_comparison.dat")
    ml_forces = np.loadtxt(root / "MLIP-Forces_comparison.dat")
    ml_stress = np.loadtxt(root / "MLIP-Stress_comparison.dat")

    assert ml_energy.shape == (nconfs + nconfs_init, 2)
    assert ml_forces.shape == ((nconfs + nconfs_init) * natoms * 3, 2)
    assert ml_stress.shape == ((nconfs + nconfs_init) * 6, 2)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .calc_manager import CalcManager
from .abinit_manager import AbinitManager
from .dlm_calc import DlmCalcManager
from .database_calc import DatabaseCalc

__all__ = ["CalcManager",
           "DlmCalcManager",
           "AbinitManager",
           "DatabaseCalc",
           ]



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .mlip_manager import MlipManager
from .descriptor import Descriptor, SumDescriptor
from .mliap_descriptor import MliapDescriptor
from .snap_descriptor import SnapDescriptor
from .mtp_model import MomentTensorPotential
from .linear_potential import LinearPotential
from .delta_learning import DeltaLearningPotential
from .mbar_manager import MbarManager
from .ace_descriptor import AceDescriptor
from .tensorpotential import TensorpotPotential
from .weights import (UniformWeight,
                      IncreasingWeight,
                      EnergyBasedWeight,
                      FixedWeight)
__all__ = ['MlipManager',
           'Descriptor',
           'SumDescriptor',
           'MliapDescriptor',
           'SnapDescriptor',
           'LinearPotential',
           'MomentTensorPotential',
           'DeltaLearningPotential',
           'MbarManager',
           'UniformWeight',
           'EnergyBasedWeight',
           'FixedWeight',
           'AceDescriptor',
           'TensorpotPotential',
           'IncreasingWeight',]



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import shutil
from ase.build import bulk
from ... import context  # noqa

from mlacs.state import LammpsState


def create_mlip():
    mlip_model = """# Cu MLIP parameters
# Descriptor   snap

# nelems   ncoefs
1 15
  -0.209698103962831089441820608954
  -0.002981838238410259092625587130
   0.030446044924109540857903866140
  -0.087136152276158285179263884856
   0.026362053083803682618224684120
  -0.100902848724669771951489849471
   0.038311487921084796948179018727
  -0.064435647545731972174465340686
   0.001572805371360150132609145857
  -0.032502969114244058268514692145
  -0.115870519655493775523780186631
  -0.049361552677329688376683947126
   0.000011351286669566577674714281
  -0.023482718193700437281368209597
   0.004556622739442980560276019730"""
    mlip_descriptor = """# Cu MLIP parameters
# Descriptor:  snap
# Model:       linear

rcutfac         4.2
twojmax         4
rfac0           0.99363
rmin0           0.0
switchflag      1
bzeroflag       1
wselfallflag    0



nelems      1
elems       Cu
radelems    0.5
welems      1.0"""
    root = Path()
    path = root / "MLIP"
    path.mkdir(parents=True, exist_ok=True)
    model_path = path / 'MLIP.model'
    descriptor_path = path / 'MLIP.descriptor'
    model_path.write_text(mlip_model)
    descriptor_path.write_text(mlip_descriptor)
    return model_path.absolute(), descriptor_path.absolute()


def cleanup():
    root = Path()
    paths = [Path("MLIP"), Path("MolecularDynamics"), Path("Trajectory")]
    for path in paths:
        if (root/path).exists():
            shutil.rmtree(root/path)


def test_NVT():
    """
    Make sure NVT correctly add atoms.info['info_state']
    """
    cleanup()
    mpath, dpath = create_mlip()

    ls = LammpsState(273.15, nsteps=100, nsteps_eq=20, dt=0.5)
    at = bulk("Cu").repeat(2)
    ps = f"mliap model linear {mpath} descriptor sna {dpath}"
    pc = ['* * Cu']
    model_post = None
    atom_style = "atomic"
    eq = False

    md_at = ls.run_dynamics(at, ps, pc, model_post, atom_style, eq)

    # Test
    assert ('info_state' in md_at.info)
    info = md_at.info['info_state']
    assert (info['ensemble'] == "NVT")
    assert (info['temperature'] == 273.15)
    assert (info['pressure'] is None)
    assert ('volume' not in md_at.info)
    cleanup()


def test_NPT():
    """
    Make sure NPT correctly add atoms.info['info_state']
    """
    cleanup()
    mpath, dpath = create_mlip()

    ls = LammpsState(3200, nsteps=100, nsteps_eq=20, dt=0.5, pressure=1)
    at = bulk("Cu").repeat(2)
    ps = f"mliap model linear {mpath} descriptor sna {dpath}"
    pc = ['* * Cu']
    model_post = None
    atom_style = "atomic"
    eq = False

    md_at = ls.run_dynamics(at, ps, pc, model_post, atom_style, eq)

    # Test
    assert ('info_state' in md_at.info)
    info = md_at.info['info_state']
    assert (info['ensemble'] == "NPT")
    assert (info['temperature'] == 3200)
    assert (info['pressure'] == 1)
    assert ('volume' not in md_at.info)
    cleanup()


def test_tstop():
    """
    Make sure NPT correctly add atoms.info['info_state']
    """
    cleanup()
    mpath, dpath = create_mlip()

    ls = LammpsState(300, nsteps=100, nsteps_eq=20,
                     dt=0.5, pressure=1, t_stop=500)
    at = bulk("Cu").repeat(2)
    ps = f"mliap model linear {mpath} descriptor sna {dpath}"
    pc = ['* * Cu']
    model_post = None
    atom_style = "atomic"
    eq = False

    md_at = ls.run_dynamics(at, ps, pc, model_post, atom_style, eq)

    # Test
    assert ('info_state' in md_at.info)
    info = md_at.info['info_state']
    assert (info['ensemble'] == "NPT")
    assert (info['temperature'] is not None)
    assert (info['pressure'] == 1)
    assert ('volume' not in md_at.info)
    cleanup()


def test_pstop():
    """
    Make sure NPT correctly add atoms.info['info_state']
    """
    cleanup()
    mpath, dpath = create_mlip()

    ls = LammpsState(3200, nsteps=10, nsteps_eq=10,
                     dt=0.5, pressure=1, p_stop=10)
    at = bulk("Cu").repeat(2)
    ps = f"mliap model linear {mpath} descriptor sna {dpath}"
    pc = ['* * Cu']
    model_post = None
    atom_style = "atomic"
    eq = False

    md_at = ls.run_dynamics(at, ps, pc, model_post, atom_style, eq)

    # Test
    assert ('info_state' in md_at.info)
    info = md_at.info['info_state']
    assert (info['ensemble'] == "NPT")
    assert (info['temperature'] == 3200)
    assert (info['pressure'] is not None)
    assert ('volume' not in md_at.info)
    cleanup()



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk
from ase.calculators.emt import EMT

from mlacs.mlip import SnapDescriptor, LinearPotential
from mlacs.state import LammpsState
from mlacs import OtfMlacs


"""
Example of a MLACS simulation of Cu at several temperature
The true potential is the EMT as implemented in ASE
"""


@pytest.fixture
def expected_folder(expected_folder_base):
    return expected_folder_base


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


def test_snap_restart(root, treelink):
    # Parameters --------------------------------------------------------------
    temperature = 300  # K
    nconfs = 2
    nsteps = 10
    nsteps_eq = 10
    neq = 1
    cell_size = 2
    rcut = 4.2
    dt = 1.5  # fs
    mlip_params = {"twojmax": 4}

    # Supercell creation ------------------------------------------------------
    atoms = bulk('Cu', cubic=True).repeat(cell_size)
    calc = EMT()

    # Running Initial Lammps State
    tmp_state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                            nsteps_eq=nsteps_eq)
    tmp_desc = SnapDescriptor(atoms, rcut, mlip_params)
    tmp_mlip = LinearPotential(tmp_desc, folder="Snap")
    sampling = OtfMlacs(atoms, tmp_state, calc, tmp_mlip, neq=neq,
                        keep_tmp_mlip=False)
    sampling.run(nconfs)

    # Running Restart Lammps State
    tmp_state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                            nsteps_eq=nsteps_eq)
    tmp_desc = SnapDescriptor(atoms, rcut, mlip_params)
    tmp_mlip = LinearPotential(tmp_desc, folder="Snap")
    sampling = OtfMlacs(atoms, tmp_state, calc, tmp_mlip, neq=neq,
                        keep_tmp_mlip=False)
    sampling.run(nconfs)
    assert sum(sampling.nconfs) == 4



"""
// Copyright (C) 2022-2024 MLACS group (PR, AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.units import kB, GPa

from ..core.manager import Manager
from ..utilities.thermo import (free_energy_harmonic_oscillator,
                                free_energy_com_harmonic_oscillator)
from ..utilities.io_lammps import get_msd_input
from ..utilities.io_lammps import LammpsBlockInput

from ..state.lammps_state import LammpsState
from .thermostate import ThermoState


# ========================================================================== #
# ========================================================================== #
class EinsteinSolidState(ThermoState):
    """
    Class for performing thermodynamic integration from
    an Einstein crystal reference

    Parameters
    ----------
    atoms: :class:`ase.Atoms`
        ASE atoms object on which the simulation will be performed

    pair_style: :class:`str`
        pair_style for the LAMMPS input

    pair_coeff: :class:`str` or :class:`list` of :class:`str`
        pair_coeff for the LAMMPS input

    temperature: :class:`float`
        Temperature of the simulation

    pressure: :class:`float`
        Pressure. None default value

    fcorr1: :class:`float` or ``None``
        First order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    fcorr2: :class:`float` or ``None``
        Second order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    k: :class:`float` or :class:`list` of :class:float` or ``None``
        Spring constant for the Einstein crystal reference.
        If a float, all atoms type have the same spring constant.
        If a list, a value for each atoms type should be provided.
        If ``None``, a short simulation is run to determine the optimal value.
        Default ``None``

    dt: :class:`int` (optional)
        Timestep for the simulations, in fs. Default ``1``

    damp : :class:`float` (optional)
        Damping parameter.
        If ``None``, a damping parameter of  1000 x dt is used.

    pdamp: :class:`float` (optional)
        Pressure damping parameter, used is the pressure is not `None`
        By default, this correspond to 1000 times the timestep.

    nsteps: :class:`int` (optional)
        Number of production steps. Default ``10000``.

    nsteps_eq: :class:`int` (optional)
        Number of equilibration steps. Default ``5000``.

    nsteps_md: :class:`int` (optional)
        Number of steps used to compute the spring constants.
        Default `25000`

    nsteps_averaging: :class:`int` (optional)
        Number of step for equilibrate ideal structure
        at zero or finite pressure.
        Default ``10000``.

    rng: :class:`RNG object`
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    langevin: :class:`bool`
        Whether to use a langevin thermostat. Default `True`

    logfile : :class:`Bool` (optional)
        Activate file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Activate Name of the file for saving the MLMD trajectory dump.
        If ``None``, no dump file is created. Default ``None``.

    interval : :class:`int` (optional)
        Number of steps between log and traj writing. Override
        loginterval. Default ``50``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.
    """
    def __init__(self,
                 atoms,
                 pair_style,
                 pair_coeff,
                 temperature,
                 pressure=None,
                 fcorr1=None,
                 fcorr2=None,
                 k=None,
                 dt=1,
                 damp=None,
                 pdamp=None,
                 nsteps=50000,
                 nsteps_eq=5000,
                 nsteps_msd=25000,
                 nsteps_averaging=10000,
                 rng=None,
                 langevin=True,
                 logfile=None,
                 trajfile=None,
                 interval=500,
                 loginterval=50,
                 **kwargs):

        kwargs.setdefault('folder', f"Solid_T{temperature}K")

        super().__init__(atoms,
                         pair_style,
                         pair_coeff,
                         dt=dt,
                         nsteps=nsteps,
                         nsteps_eq=nsteps_eq,
                         nsteps_averaging=nsteps_averaging,
                         rng=rng,
                         langevin=langevin,
                         logfile=logfile,
                         trajfile=trajfile,
                         interval=interval,
                         loginterval=loginterval,
                         **kwargs)

        self.atoms = atoms
        self.temperature = temperature
        self.pressure = pressure
        self.damp = damp
        self.pdamp = pdamp
        self.nsteps_msd = nsteps_msd
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()
        self.langevin = langevin

        self.fcorr1 = fcorr1
        self.fcorr2 = fcorr2

        if self.pressure is not None:
            self.equilibrate = True
        else:
            self.equilibrate = False

        self.k = k
        if self.k is not None:
            if isinstance(self.k, list):
                if not len(self.k) == len(self.elem):
                    msg = "The spring constant paramater has to be a " + \
                          "float or a list of length n=number of " + \
                          "different species in the system"
                    raise ValueError(msg)
            elif isinstance(self.k, (float, int)):
                self.k = [self.k] * len(self.elem)
            else:
                msg = "The spring constant parameter k has to be a " + \
                      "float or a list of length n=\'number of " + \
                      "different species in the system\'"
                raise ValueError(msg)

        self.dt = dt

# ========================================================================== #
    @Manager.exec_from_path
    def run(self):
        """
        """
        if self.equilibrate:
            self.run_averaging()

        if self.k is None:
            # First get optimal spring constant
            self.compute_msd()

        self.run_dynamics(self.atoms, self.pair_style, self.pair_coeff)

# ========================================================================== #
    @Manager.exec_from_path
    def compute_msd(self):
        """
        """
        msd_state = LammpsState(self.temperature,
                                pressure=self.pressure,
                                nsteps=self.nsteps_msd,
                                nsteps_eq=self.nsteps_eq,
                                lammpsfname='lammps_input_msd.in',
                                blocks=get_msd_input(self, 'msd.dat'),
                                workdir=self.workdir,
                                folder=self.folder,
                                subfolder='MSD',
                                rng=self.rng)

        msd_state.run_dynamics(self.atoms,
                               self.pair_style,
                               self.pair_coeff)

        kall = []
        with open("msd.dat", "w") as f:
            for e in self.elem:
                data = np.loadtxt(str(msd_state.path / f"msd{e}.dat"))
                nat = np.count_nonzero([a == e for a in
                                        self.atoms.get_chemical_symbols()])
                k = 3 * kB * self.temperature / data.mean()
                kall.append(k)
                f.write(e + " {0}   {1:10.5f}\n".format(nat, k))
        self.k = kall

# ========================================================================== #
    @Manager.exec_from_path
    def postprocess(self):
        """
        Compute the free energy from the simulation
        """
        # Get needed value/constants
        vol = self.atoms.get_volume()
        nat_tot = len(self.atoms)

        # Compute some oscillator frequencies and number
        # of atoms for each species
        omega = []
        nat = []
        for iel, e in enumerate(self.elem):
            omega.append(np.sqrt(self.k[iel] / (self.masses[iel])))
            nat.append(np.count_nonzero([a == e for a in
                                         self.atoms.get_chemical_symbols()]))

        # Compute free energy of the Einstein crystal
        f_harm = free_energy_harmonic_oscillator(omega,
                                                 self.temperature,
                                                 nat)  # eV/at

        # Compute the center of mass correction
        f_cm = free_energy_com_harmonic_oscillator(self.k,
                                                   self.temperature,
                                                   nat,
                                                   vol,
                                                   self.masses)  # eV/at

        # Compute the work between einstein crystal and the MLIP
        dE_f, lambda_f = np.loadtxt("forward.dat", unpack=True)
        dE_b, lambda_b = np.loadtxt("backward.dat", unpack=True)
        int_f = np.trapz(dE_f, lambda_f)
        int_b = np.trapz(dE_b, lambda_b)

        work = (int_f - int_b) / 2.0

        free_energy = f_harm + f_cm + work
        free_energy_corrected = free_energy
        if self.fcorr1 is not None:
            free_energy_corrected += self.fcorr1
        if self.fcorr2 is not None:
            free_energy_corrected += self.fcorr2

        if self.pressure is not None:
            pv = self.pressure*GPa*vol/nat_tot
        else:
            pv = 0.0
        with open("free_energy.dat", "w") as f:
            header = "#   T [K]     Fe tot [eV/at]     " + \
                     "Fe harm [eV/at]      Work [eV/at]     " + \
                     "Fe com [eV/at]      PV [eV/at]"
            results = f"{self.temperature:10.3f}     " + \
                      f"{free_energy:10.6f}         " + \
                      f"{f_harm:10.6f}          " + \
                      f"{work:10.6f}         " + \
                      f"{f_cm:10.6f}         " + \
                      f"{pv:10.6f}"
            if self.fcorr1 is not None:
                header += "    Delta F1 [eV/at]"
                results += f"         {self.fcorr1:10.6f}"
            if self.fcorr2 is not None:
                header += "    Delta F2 [eV/at]"
                results += f"          {self.fcorr2:10.6f}"
            if self.fcorr1 is not None or self.fcorr2 is not None:
                header += "    Fe corrected [eV/at]"
                results += f"           {free_energy_corrected:10.6f}"
            header += "\n"
            f.write(header)
            f.write(results)

        msg = "Summary of results for this state\n"
        msg += '============================================================\n'
        msg += "Frenkel-Ladd path integration, with an " + \
               "Einstein crystal reference\n"
        msg += "Reference Einstein crystal spring :\n"
        for iel, e in enumerate(self.elem):
            msg += f"    For {e} :                   " + \
                   f"{self.k[iel]:10.6f} eV/angs^2\n"
        msg += f"Temperature :                   {self.temperature:10.3f} K\n"
        msg += f"Volume :                        {vol/nat_tot:10.3f} angs^3\n"
        msg += f"Free energy :                   {free_energy:10.6f} eV/at\n"
        msg += f"Excess work :                   {work:10.6f} eV/at\n"
        msg += f"Einstein crystal free energy :  {f_harm:10.6f} eV/at\n"
        msg += f"Center of mass free energy :    {f_cm:10.6f} eV/at\n"
        if self.fcorr1 is not None:
            msg += "1st order true pot correction : " + \
                   f"{self.fcorr1:10.6f} eV/at\n"
        if self.fcorr2 is not None:
            msg += "2nd order true pot correction : " + \
                   f"{self.fcorr2:10.6f} eV/at\n"
        if self.fcorr1 is not None or self.fcorr2 is not None:
            msg += "Free energy corrected :         " + \
                   f"{free_energy_corrected:10.6f} eV/at\n"
        # add Fe or Fe_corrected to return to be read for cv purpose and RS
        if self.fcorr1 is not None or self.fcorr2 is not None:
            free_energy = free_energy_corrected
        if self.pressure:
            free_energy += pv
        return msg, free_energy

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """

        """
        if self.damp is None:
            self.damp = "$(100*dt)"

        temp = self.temperature
        self.info_dynamics["temperature"] = temp
        if self.langevin:
            langevinseed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("thermostat", "Integrators")
        block("timestep", f"timestep {self.dt / 1000}")
        block("momenta", "velocity all create " +
              f"{temp} {langevinseed} dist gaussian")
        # If we are using Langevin, we want to remove the random part
        # of the forces
        if self.langevin:
            block("rmv_langevin", "fix ff all store/force")
            txt = f"fix f1 all langevin {temp} {temp} {self.damp} " + \
                  f"{langevinseed} zero yes"
            block("langevin", txt)
            block("nve", "fix f2 all nve")
        else:
            block("nvt", f"fix f1 all nvt temp {temp} {temp} {self.damp}")
        block("compute temp without cm", "compute c1 all temp/com")
        block("fix cm", "fix_modify f1 temp c1")
        return block

# ========================================================================== #
    def _get_block_traj(self, el):
        """
        """
        if self.trajfile:
            block = LammpsBlockInput("dump", "Dumping")
            txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} "
            txt += "id type xu yu zu vx vy vz fx fy fz "
            txt += "element"
            block("dump", txt)
            block("dump_modify1", "dump_modify dum1 append yes")
            txt = "dump_modify dum1 element " + " ".join([p for p in el])
            block("dump_modify2", txt)
            return block
        else:
            pass

# ========================================================================== #
    def _get_neti(self):
        """
        Write the LAMMPS input neti in solids
        """
        blocks = []
        block0 = LammpsBlockInput("einstein params",
                                  "Integrators and variables for solid neti")
        for iel, el in enumerate(self.elem):
            txt = f"fix ff{el} {el} ti/spring " + \
                  f"{self.k[iel]} {self.nsteps} {self.nsteps_eq} " + \
                  "function 2"
            block0(f"ti_spring_{el}", txt)
            block0("dE", f"variable dE equal (pe-f_ff{el})/atoms")
        block0("lambda", f"variable lambda equal f_ff{self.elem[0]}[1]")
        blocks.append(block0)

        block1 = LammpsBlockInput("eq fwd",
                                  "Equilibration without Einstein potential")
        block1("run eq fwd", f"run {self.nsteps_eq}")
        blocks.append(block1)

        block2 = LammpsBlockInput("fwd", "Forward Integration")
        block2("write fwd", "fix f4 all print 1 \"${dE} ${lambda}\" " +
                            "screen no append forward.dat title " +
                            "\"# pe  lambda\"")
        block2("run fwd", f"run {self.nsteps}")
        block2("unifx write fwd", "unfix f4")
        blocks.append(block2)

        block3 = LammpsBlockInput("eq bwd",
                                  "Equilibration with only Einstein potential")
        block3("run eq bwd", f"run {self.nsteps_eq}")
        blocks.append(block3)

        block4 = LammpsBlockInput("bwd", "Backward Integration")
        block4("write bwd", "fix f4 all print 1 \"${dE} ${lambda}\" " +
                            "screen no append backward.dat title " +
                            "\"# pe  lambda\"")
        blocks.append(block4)

        return blocks

        block4 = LammpsBlockInput("bwd", "Backward Integration")
        block4("write bwd", "fix f4 all print 1 \"${dE} ${lambda}\" " +
               "screen no append backward.dat title \"# pe  lambda\"")
        blocks.append(block4)

        return blocks

        block4 = LammpsBlockInput("bwd", "Backward Integration")
        block4("write bwd", "fix f4 all print 1 \"${dE} ${lambda}\" " +
               "screen no append backward.dat title \"# pe  lambda\"")
        blocks.append(block4)

        return blocks

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        if self.damp is None:
            damp = 100 * self.dt

        msg = "Thermodynamic Integration using Frenkel-Ladd " + \
              "path and an Einstein crystal\n"
        msg += f"Temperature :                   {self.temperature}\n"
        msg += f"Langevin damping :              {damp} fs\n"
        msg += f"Timestep :                      {self.dt} fs\n"
        msg += f"Number of steps :               {self.nsteps}\n"
        msg += f"Number of equilibration steps : {self.nsteps_eq}\n"
        if self.k is None:
            msg += "Reference einstein crystal to be computed\n"
        else:
            msg += "Reference einstein crystal spring :\n"
            for iel, e in enumerate(self.elem):
                msg += f"    For {e} :                   " + \
                       f"k = {self.k[iel]} eV/angs^2\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .mlas import Mlas
from .core import Manager
from .properties import PropertyManager


# ========================================================================== #
# ========================================================================== #
class OtfMlacs(Mlas, Manager):
    r"""
    A Learn on-the-fly simulation constructed in order to sample approximate
    distribution

    Parameters
    ----------

    atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms`
        the atom object on which the simulation is run.

    state: :class:`StateManager` or :class:`list` of :class:`StateManager`
        Object determining the state to be sampled

    calc: :class:`ase.calculators` or :class:`CalcManager`
        Class controlling the potential energy of the system
        to be approximated.
        If a :class:`ase.calculators` is attached, the :class:`CalcManager`
        is automatically created.

    mlip: :class:`MlipManager` (optional)
        Object managing the MLIP to approximate the real distribution
        Default is a LammpsMlip object with a snap descriptor,
        ``5.0`` angstrom rcut with ``8`` twojmax.

    neq: :class:`int` (optional)
        The number of equilibration iteration. Default ``10``.

    workdir: :class:`str` (optional)
        The directory in which to run the calculation.

    confs_init: :class:`int` or :class:`list` of :class:`ase.Atoms` (optional)
        If :class:`int`, Number of configurations used to train a preliminary
        MLIP. The configurations are created by rattling the first structure.
        If :class:`list` of :class:`ase.Atoms`, The atoms that are to be
        computed in order to create the initial training configurations.
        Default ``None``.

    std_init: :class:`float` (optional)
        Variance (in :math:`\mathring{a}^2`) of the displacement
        when creating initial configurations.
        Default :math:`0.05 \mathring{a}^2`

    keep_tmp_mlip: :class:`Bool` (optional)
        Keep every generated MLIP. If True and using MBAR, a restart will
        recalculate every previous MLIP.weight using the old coefficients.
        Default ``True``.

    prefix: :class:`str` (optional)
        The prefix to prepend the name of the States files.

    ncprefix: :class:`str` (optional)
        The prefix to prepend the name of the *HIST.nc file.
        Script name format: ncprefix + scriptname + '_HIST.nc'.
        Default `''`.

    ncformat: :class:`str` (optional)
        The format of the *HIST.nc file. One of the five flavors of netCDF
        files format available in netCDF4 python package: 'NETCDF3_CLASSIC',
        'NETCDF3_64BIT_OFFSET', 'NETCDF3_64BIT_DATA','NETCDF4_CLASSIC',
        'NETCDF4'.
        Default ``NETCDF3_CLASSIC``.
    """

    def __init__(self,
                 atoms,
                 state,
                 calc,
                 mlip=None,
                 prop=None,
                 neq=10,
                 confs_init=None,
                 std_init=0.05,
                 keep_tmp_mlip=True,
                 workdir='',
                 prefix='Trajectory',
                 ncprefix='',
                 ncformat='NETCDF3_CLASSIC'):
        Mlas.__init__(self, atoms, state, calc, mlip=mlip, prop=None, neq=neq,
                      confs_init=confs_init, std_init=std_init,
                      keep_tmp_mlip=keep_tmp_mlip, workdir=workdir,
                      prefix=prefix, ncprefix=ncprefix, ncformat=ncformat)

        # RB: Move the initialization of properties out of Mlas.
        self._initialize_properties(prop)

# ========================================================================== #
    def _initialize_properties(self, prop):
        """Create property object"""
        self.prop = PropertyManager(prop)

        if self.ncfile is not None:
            if not self.launched:
                self.ncfile.create_nc_var(self.prop.manager)

            self.prop.workdir = self.workdir
            # RB: I think this is not necessary anymore
            # if not self.prop.folder:
            #     self.prop.folder = 'Properties'

            self.prop.isfirstlaunched = not self.launched
            self.prop.ncfile = self.ncfile

# ========================================================================== #
    def _compute_properties(self):
        """
        Main method to compute/save properties of OtfMlacs objects.
        """

        if self.prop.manager is not None:
            self.prop.calc_initialize(atoms=self.atoms)
            msg = self.prop.run(self.step)
            self.log.logger_log.info(msg)
            self.prop.save_prop(self.step)
            if self.prop.check_criterion:
                msg = "All property calculations are converged, " + \
                      "stopping MLACS ...\n"
                self.log.logger_log.info(msg)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk
from ase.calculators.emt import EMT
from mlacs.mlip import DeltaLearningPotential, LinearPotential
from mlacs.mlip import MliapDescriptor

from mlacs.state import LammpsState
from mlacs import OtfMlacs


"""
Example of a MLACS simulation of Cu at several temperature
The true potential is the EMT as implemented in ASE
"""


@pytest.fixture
def expected_folder(expected_folder_base):
    return expected_folder_base


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


def test_zbl_restart(root, treelink):
    # Parameters --------------------------------------------------------------
    temperature = 300  # K
    nconfs = 2
    nsteps = 10
    nsteps_eq = 10
    neq = 1
    cell_size = 2
    rcut = 4.2
    dt = 1.5  # fs
    mlip_params = {"twojmax": 4}

    # Supercell creation ------------------------------------------------------
    atoms = bulk('Cu', cubic=True).repeat(cell_size)
    calc = EMT()

    # Running Initial Lammps State
    state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                        nsteps_eq=nsteps_eq)

    zbl_style = ["zbl 0.9 2.0"]
    zbl_coeff = ['1 1 29 29']
    desc = MliapDescriptor(atoms, rcut, mlip_params, style="snap")
    mlip = LinearPotential(desc, folder="Snap")
    dlpot = DeltaLearningPotential(model=mlip,
                                   pair_style=zbl_style,
                                   pair_coeff=zbl_coeff,
                                   atom_style="atomic")

    sampling = OtfMlacs(atoms, state, calc, dlpot, neq=neq)
    sampling.run(nconfs)

    # Running Restart Lammps State
    state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                        nsteps_eq=nsteps_eq)
    desc = MliapDescriptor(atoms, rcut, mlip_params, style="snap")
    mlip = LinearPotential(desc, folder="Snap")
    dlpot = DeltaLearningPotential(model=mlip,
                                   pair_style=zbl_style,
                                   pair_coeff=zbl_coeff,
                                   atom_style="atomic")

    sampling = OtfMlacs(atoms, state, calc, dlpot, neq=neq)
    sampling.run(nconfs)
    assert sum(sampling.nconfs) == 4

# def test_mtp_restart(root, treelink):
#    pass



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk
from ase.calculators.emt import EMT

from mlacs.mlip import MliapDescriptor, LinearPotential
from mlacs.state import LangevinState
from mlacs import OtfMlacs


"""
Example of a MLACS simulation of Cu at several temperature
The true potential is the EMT as implemented in ASE
"""


@pytest.fixture
def expected_folder(expected_folder_base):
    return ["Snap"]


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


def test_langevin_mliap_restart(root, treelink):
    # Parameters --------------------------------------------------------------
    temperature = 300  # K
    nconfs = 2
    nsteps = 10
    nsteps_eq = 10
    neq = 1
    cell_size = 2
    rcut = 4.2
    dt = 1.5  # fs
    friction = 0.01
    mlip_params = {"twojmax": 4}

    # Supercell creation  -----------------------------------------------------
    atoms = bulk('Cu', cubic=True).repeat(cell_size)
    calc = EMT()

    # Prepare the On The Fly Machine-Learning Assisted Sampling simulation ----
    # Creation of the MLIP Manager
    mlip_params = dict(twojmax=4)

    # Running Initial Langevin State
    tmp_state = LangevinState(temperature, dt=dt, nsteps=nsteps,
                              nsteps_eq=nsteps_eq, friction=friction)
    tmp_desc = MliapDescriptor(atoms, rcut, mlip_params, style="snap")
    tmp_mlip = LinearPotential(tmp_desc, folder="Snap")
    sampling = OtfMlacs(atoms, tmp_state, calc, tmp_mlip, neq=neq)
    sampling.run(nconfs)

    # Running Restart Langevin State
    tmp_state = LangevinState(temperature, dt=dt, nsteps=nsteps,
                              nsteps_eq=nsteps_eq, friction=friction)
    tmp_desc = MliapDescriptor(atoms, rcut, mlip_params, style="snap")
    tmp_mlip = LinearPotential(tmp_desc, folder="Snap")
    sampling = OtfMlacs(atoms, tmp_state, calc, tmp_mlip, neq=neq)
    sampling.run(nconfs)
    assert sum(sampling.nconfs) == 4



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import os

import numpy as np
from scipy.stats import gaussian_kde
import matplotlib as mpl
import matplotlib.pyplot as plt

from . import compute_correlation
from mlacs.utilities.io_abinit import MlacsHist
from mlacs.utilities.units import unit_converter

cyan = "#17becf"
blue = "#1f77b4"
red = "#d62728"
orange = "#ff7f0e"
green = "#2ca02c"
violet = "#9467bd"
grey = "#7f7f7f"

colors = [blue, red, orange, green, violet, cyan, grey]


def plot_correlation(ax,
                     data,
                     color=blue,
                     marker="o",
                     datatype=None,
                     density=False,
                     weight=None,
                     cmap="inferno",
                     showrmse=True,
                     showmae=True,
                     showrsquared=True,
                     size=5,
                     axcbar=None):
    """
    Function to plot the correlation between true and model data on an axes

    Parameters:
    -----------
    ax: Axes.axes
        The axes on which to plot the data
    data: `np.ndarray`
        The data to plot. Has to be of shape (n, 2)
        with n the number of datapoint.
    color:
        The color of the marker in the scatter plot.
        Ignored if density is True.
    marker:
        Marker type in the scatter plot.
    datatype: `None` or `str`
        The type of data. Can be either "energy", "forces" or "stress"
    density: `Bool`
        If True, each datapoint is colored according to the density
        of data
    cmap: `str`
        The colormap used if density is True.
        Ignored if density is False
    showrmse: `Bool`
        Whether to show the RMSE on the plot
    showmae: `Bool`
        Whether to show the MAE on the plot
    showrsquared: `Bool`
        Whether to show the R^2 on the plot

    Returns:
    --------
    ax
    """

    if datatype == "energy":
        data[:, 1] -= data[:, 0].min()
        data[:, 0] -= data[:, 0].min()

    cancbar = np.any([weight is not None, density])
    if axcbar is not None and not cancbar:
        msg = "You need weight or density to use plot a color bar"
        raise ValueError(msg)

    datatrue = data[:, 0]
    datatest = data[:, 1]

    mindata = data.min()
    maxdata = data.max()
    minmax = [mindata, maxdata]

    rmse, mae, rsquared = compute_correlation(data, weight)

    if density:
        xy = np.vstack([datatrue, datatest])
        z = gaussian_kde(xy)(xy)
        norm = mpl.colors.LogNorm(z.min(), z.max())
        idx = z.argsort()
        plot = ax.scatter(datatrue[idx], datatest[idx], c=z[idx],
                          linewidths=5, norm=norm, s=size, cmap=cmap)
        if axcbar is not None:
            mpl.colorbar.Colorbar(axcbar, plot, cmap=cmap, norm=norm)
            axcbar.set_ylabel("Density")

    elif weight is not None:
        if datatype != "energy":
            w = []
            if len(datatrue) % len(weight) == 0:
                n = int(len(datatrue)/len(weight))
                for i, _w in enumerate(weight):
                    w.extend(_w * np.ones(n) / n)
            else:
                msg = "Number of weight not consistent with the Database"
                raise ValueError(msg)
            weight = np.r_[w] / np.sum(np.r_[w])
        # We add a small number to the min to avoid a possible 0 with the log
        norm = mpl.colors.LogNorm(weight.min() + 1e-8, weight.max())
        plot = ax.scatter(datatrue, datatest, c=weight,
                          linewidths=5, norm=norm, s=size, cmap=cmap)
        if axcbar is not None:
            mpl.colorbar.Colorbar(axcbar, plot, cmap=cmap, norm=norm)
            axcbar.set_ylabel("Weight")
    else:
        ax.plot(datatrue, datatest, ls="", marker=marker,
                c=color, rasterized=True, markersize=size,
                markeredgewidth=size/5)
    ax.plot(minmax, minmax, ls="--", alpha=0.75, c=red)

    if datatype is not None:
        if datatype == "energy":
            labelx = "True energy [eV/at]"
            labely = "Model energy [eV/at]"
            unit = "[eV/at]"
        elif datatype == "forces":
            labelx = "True forces [eV/angs]"
            labely = "Model forces [eV/angs]"
            unit = "[eV/angs]"
        elif datatype == "stress":
            labelx = "True stress [GPa]"
            labely = "Model stress [GPa]"
            unit = "[GPa]"
        else:
            msg = "datVatype should be energy, forces or stress"
            raise ValueError(msg)
    else:
        labelx = None
        labely = None
        unit = ""

    if showrmse:
        ax.text(0.01, 0.9, f"RMSE = {rmse:5.4f} {unit}",
                fontsize=30,
                transform=ax.transAxes)
    if showmae:
        ax.text(0.01, 0.8, f"MAE = {mae:5.4f} {unit}",
                fontsize=30,
                transform=ax.transAxes)
    if showrsquared:
        ax.text(0.01, 0.7, f"R$^2$ = {rsquared:5.4f}",
                fontsize=30,
                transform=ax.transAxes,)

    ax.set_xlabel(labelx)
    ax.set_ylabel(labely)
    ax.set_xlim(minmax)
    ax.set_ylim(minmax)
    return ax


def plot_error(ax,
               data,
               color=blue,
               datatype=None,
               showrmse=True,
               showmae=True,
               showrsquared=True):
    """
    Function to plot the error distribution between true and model data

    Parameters:
    -----------
    ax: Axes.axes
        The axes on which to plot the data
    data: `np.ndarray`
        The data to plot. Has to be of shape (n, 2)
        with n the number of datapoint.
    color:
        The color of the marker in the scatter plot.
        Ignored if density is True.
    datatype: `None` or `str`
        The type of data. Can be either "energy", "forces" or "stress"
    showrmse: `Bool`
        Whether to show the RMSE on the plot
    showmae: `Bool`
        Whether to show the MAE on the plot
    showrsquared: `Bool`
        Whether to show the R^2 on the plot

    Returns:
    --------
    ax
    """
    dataerror = data[:, 0] - data[:, 1]

    if datatype is not None:
        if datatype == "energy":
            dataerror *= 1000
            labelx = "Energy error [meV/at]"
            unit = "[meV/at]"
        elif datatype == "forces":
            dataerror *= 1000
            labelx = "Forces error [meV/angs]"
            unit = "[meV/angs]"
        elif datatype == "stress":
            labelx = "Stress error [GPa]"
            unit = "[GPa]"
        else:
            msg = "datatype should be energy, forces or stress"
            raise ValueError(msg)
    else:
        labelx = None
        unit = ""

    errmin = -3 * dataerror.std()
    errmax = 3 * dataerror.std()
    errmean = dataerror.mean()
    minmax = [errmin - errmean, errmean + errmax]

    rmse, mae, rsquared = compute_correlation(data)

    kdeerror = gaussian_kde(dataerror)

    x = np.linspace(errmin, errmax, 1000)
    kde_pred = kdeerror(x)
    kde_pred *= 100 / (kde_pred).sum()
    ax.axvline(errmean, c=grey, ls="--")
    ax.plot(x, kde_pred, c="k")
    ax.fill_between(x, kde_pred, alpha=0.75)

    if showrmse:
        ax.text(0.01, 0.9, f"RMSE = {rmse:5.4f} {unit}",
                fontsize=30,
                transform=ax.transAxes)
    if showmae:
        ax.text(0.01, 0.8, f"MAE = {mae:5.4f} {unit}",
                fontsize=30,
                transform=ax.transAxes)
    if showrsquared:
        ax.text(0.01, 0.7, f"R$^2$ = {rsquared:5.4f}",
                fontsize=30,
                transform=ax.transAxes,)

    ax.set_xlabel(labelx)
    ax.set_ylabel("Density [%]")
    ax.set_xlim(minmax)
    ax.set_ylim(0)
    return ax


def plot_weights(ax, weights, color=blue, fontsize=30):
    """
    Function to plot the error distribution between true and model data

    Parameters:
    -----------
    ax: Axes.axes
        The axes on which to plot the data
    weights: `np.ndarray`
        The weights to plot.
    color:
        The color of the marker in the scatter plot.
        Ignored if density is True.

    Returns:
    --------
    ax
    """
    xrange = np.arange(len(weights))
    neff = np.sum(weights)**2 / np.sum(weights**2)

    ax.bar(xrange, weights)
    ax.text(0.01, 0.9, f"Eff. N. conf = {neff:5.4f}",
            transform=ax.transAxes)
    ax.set_ylim(0)
    ax.set_xlabel("Configuration index")
    ax.set_ylabel("Weight")
    return ax


def init_rcParams():
    """
    """
    mpl.rcParams["lines.linewidth"] = 5
    mpl.rcParams["lines.markeredgecolor"] = "k"
    mpl.rcParams["lines.markersize"] = 25
    mpl.rcParams["lines.markeredgewidth"] = 5

    mpl.rcParams['figure.dpi'] = 300

    mpl.rcParams["font.size"] = 30

    mpl.rcParams["axes.linewidth"] = 5

    mpl.rcParams["xtick.top"] = True
    mpl.rcParams["xtick.major.size"] = 12
    mpl.rcParams["xtick.major.width"] = 5
    mpl.rcParams["xtick.direction"] = "in"

    mpl.rcParams["ytick.right"] = True
    mpl.rcParams["ytick.major.size"] = 12
    mpl.rcParams["ytick.major.width"] = 5
    mpl.rcParams["ytick.direction"] = "in"


# ========================================================================== #
# ========================================================================== #
class HistPlot:
    """
    Class to handle the plots of Abinit-like *HIST.nc file.

    Parameters
    ----------

    ncpath: :class:`str` or :class:`Path` of `pathlib` module (optional)
        Absolute path to *HIST.nc file, i.e. `path_to_ncfile/ncfilename`.
        Default ''.
    """

    def __init__(self,
                 ncpath=''):

        mpl.rcdefaults()
        mpl.rcParams["font.size"] = 10
        mpl.rcParams['figure.dpi'] = 300

        if os.path.isfile(ncpath):
            ncfile = MlacsHist(ncpath=ncpath)
            dict_var_units = ncfile.get_units()
            var_dim_dict = ncfile.var_dim_dict
            dict_name_label = {x[0]: lab for lab, x in var_dim_dict.items()}
            dict_name_label['press'] = 'Pressure'
            self.ncfile = ncfile
            self.dict_name_label = dict_name_label
            self.dict_var_units = dict_var_units
            self.basic_obs = ['temper', 'etotal', 'press', 'vol']
            self.energy_obs = ['ekin', 'epot']

            weights_ncpath = ncpath
            if 'NETCDF3' in ncfile.ncformat:
                weights_ncpath = ncpath.replace('HIST', 'WEIGHTS')
            self.weights_ncfile = MlacsHist(ncpath=weights_ncpath)
        else:
            msg = '*HIST.nc file not found.'
            raise FileNotFoundError(msg)

# ========================================================================== #
    def _initialize_weights(self, get_dict_weights=False):
        """
        Initialize attributes that are relevant when weights are needed.
        """
        weights_ncfile = self.weights_ncfile
        weights = weights_ncfile.read_obs('weights')
        weights_meta = weights_ncfile.read_obs('weights_meta')
        weights_idx = weights_meta[:, 0]
        nb_effective_conf = weights_meta[:, 1][weights_idx == 1.0]
        nb_conf = weights_meta[:, 2][weights_idx == 1.0]

        self.weights = weights
        self.weights_idx = weights_idx
        self.nb_effective_conf = nb_effective_conf
        self.nb_conf = nb_conf
        self.number_of_states = weights_meta[0][-1]

        # Check if weights are uniform within tolerance
        self.uniform_weight = np.allclose(nb_conf, nb_effective_conf,
                                          atol=1e-10)

# ========================================================================== #
    def _get_dict_weights(self):
        """
        Compute a dictionary that maps MLACS iterations to weights data.
        """
        weights_idx = self.weights_idx
        weights = self.weights

        dict_weights = {}
        idx_bounds = np.argwhere(weights_idx == 1.0)[:, 0]

        for ii in range(len(idx_bounds)-1):
            iter_mlacs = ii+1
            i1, i2 = idx_bounds[ii], idx_bounds[ii+1]
            dict_weights[iter_mlacs] = [weights_idx[i1:i2], weights[i1:i2]]
        dict_weights[iter_mlacs+1] = [weights_idx[i2:], weights[i2:]]

        self.dict_weights = dict_weights
        self.available_iter = list(dict_weights.keys())

# ========================================================================== #
    def _plot_weight_distribution(self, iter_loc, ax_loc, normalize=True):
        """
        Plot the weight distribution for a specified MLACS iteration.
        """
        dict_weights = self.dict_weights
        nb_effective_conf = self.nb_effective_conf

        weights_idx = dict_weights[iter_loc][0]-1
        weights = dict_weights[iter_loc][1]

        xlabel = r"Number of configurations in database"
        ylabel = r"Weights"
        if normalize:
            weights /= np.mean(weights)
            ylabel += r"/ $ \langle $Weights$ \rangle $"

        Nconfs_loc = np.round(nb_effective_conf[iter_loc-1], 1)
        lab_str = r'$N_{\text{eff}} \simeq$'+'{}'.format(Nconfs_loc)

        ax_loc.step(weights_idx, weights, where='mid', label=lab_str,
                    zorder=10-iter_loc)
        ax_loc.set_xlabel(xlabel)
        ax_loc.set_ylabel(ylabel)

# ========================================================================== #
    def _core_plot(self, obs_name, fig=None, ax=None):
        """Plot the observable named `obs_name` in the ncfile."""
        if None in (fig, ax):
            fig, ax = plt.subplots()

        ncfile = self.ncfile
        dict_name_label = self.dict_name_label
        dict_var_units = self.dict_var_units

        try:
            nc_unit = dict_var_units[obs_name]
        except KeyError:
            msg = 'No unit found for ' + str(obs_name)
            nc_unit = ''
            raise KeyError(msg)

        # Read atomic unit observable from *HIST.nc file
        obs_au = ncfile.read_obs(obs_name)
        # Convert to custom visualization units
        observable, new_unit = unit_converter(obs_au, nc_unit, target='custom')

        obs_meta = ncfile.read_obs(obs_name + '_meta')

        state_idx = obs_meta[:, 1]
        confs_idx = np.arange(1, len(observable)+1)

        w_obs_au, w_obs_idx = ncfile.read_weighted_obs('weighted_'+obs_name)
        w_obs_data, new_unit = unit_converter(w_obs_au,
                                              nc_unit,
                                              target='custom')

        uniform_obs = np.array([np.mean(observable[:i]) for i in w_obs_idx])

        ax.plot(confs_idx, observable, label='raw data', alpha=0.7)
        ax.plot(w_obs_idx, uniform_obs, c='g', label='uniform weights',
                marker='.')

        self._initialize_weights()
        if not self.uniform_weight:
            ax.plot(w_obs_idx, w_obs_data, c='r', ls='-', label='mbar',
                    marker='.')

        ax.set_xlabel('Configuration index in database')
        obs_label = obs_name
        if obs_name in dict_name_label:
            obs_label = dict_name_label[obs_name].replace("_", " ")
        ylabel = obs_label
        if new_unit != '':
            ylabel += ' [' + new_unit + ']'
        ax.set_ylabel(ylabel)
        ax.legend(frameon=False, loc='best')
        par_title = (int(len(confs_idx)), int(max(state_idx)),)
        str_title = '# configurations: {}, # states: {}'.format(*par_title)
        fig.suptitle(str_title)

# ========================================================================== #
    def plot_thermo_basic(self, show=True, savename=''):
        """Plot the set of observables in self.basic_obs"""
        fig, ax = plt.subplots(2, 2, figsize=(9, 7))
        for idx, ax_loc in enumerate(ax.reshape(-1)):
            obs_name = self.basic_obs[idx]
            self._core_plot(obs_name, fig, ax_loc)
        fig.tight_layout()

        dynamic_savename = f"{savename or 'plot_thermo'}.pdf"
        fig.savefig(dynamic_savename, bbox_inches='tight')

        if show:
            os.system('xdg-open '+dynamic_savename)

# ========================================================================== #
    def plot_neff(self, show=True, savename='', selected_iter=None):
        """Plot Neff against Nconfs, along with some weight distributions."""
        self._initialize_weights()
        self._get_dict_weights()
        nb_conf = self.nb_conf
        nb_effective_conf = self.nb_effective_conf
        available_iter = self.available_iter
        number_of_states = self.number_of_states

        fig, ax = plt.subplots(1, 2, figsize=(7, 3))

        ax[0].plot(nb_conf, nb_effective_conf)
        ax[0].plot(nb_conf, nb_conf, c='k', ls=':', label=r'$y=x$')
        ax[0].set_xlabel('Number of configurations in database')
        ax[0].set_ylabel('Number of effective configurations')
        ax[0].set_xscale('log')
        ax[0].set_yscale('log')

        # Secondary x-axis for MLACS iterations
        def f(x): return x/number_of_states
        def g(x): return x*number_of_states
        second_x_axis_0 = ax[0].secondary_xaxis("top", functions=(f, g))
        second_x_axis_0.set_xlabel("MLACS iteration")
        ax[0].legend(frameon=False, loc=4)

        # Identify MLACS iterations to plot
        mlacs_iter_arr = [available_iter[-1]]
        if selected_iter is None:
            if len(available_iter) > 5:
                mlacs_iter_arr = np.geomspace(3, available_iter[-1], 4,
                                              dtype=int)
        else:
            if set(selected_iter) <= set(available_iter):
                selected_iter.sort()
                mlacs_iter_arr = np.array(selected_iter)
            else:
                s = sorted(set(selected_iter))
                missing_iter = [x for x in s if x not in available_iter]
                if len(missing_iter) > 0:
                    raise ValueError(
                        "The following iteration(s) are not available: "
                        f"{missing_iter}. \n"
                        f"Available iteration(s): {available_iter}")

        # Plot weight distributions and pinpoint iterations on left panel
        for iter_mlacs in mlacs_iter_arr:
            self._plot_weight_distribution(iter_mlacs, ax[1])
            ax[0].scatter(nb_conf[iter_mlacs-1],
                          nb_effective_conf[iter_mlacs-1],
                          marker='s',
                          s=20)

        ax[1].legend(frameon=False, loc='best', ncol=2)
        fig.tight_layout()

        dynamic_savename = f"{savename or 'plot_neff'}.pdf"
        fig.savefig(dynamic_savename, bbox_inches='tight')

        if show:
            os.system('xdg-open '+dynamic_savename)

# ========================================================================== #
    def plot_ith_weights(self, show=True, savename='', selected_iter=None):
        """Plot MLACS weights distribution for iteration i."""
        self._initialize_weights()
        self._get_dict_weights()
        available_iter = self.available_iter

        # Identify MLACS iterations to plot
        if selected_iter is None:
            mlacs_iter_arr = [available_iter[-1]]
        else:
            if set(selected_iter) <= set(available_iter):
                selected_iter.sort()
                mlacs_iter_arr = np.array(selected_iter)
            else:
                s = sorted(set(selected_iter))
                missing_iter = [x for x in s if x not in available_iter]
                if len(missing_iter) > 0:
                    raise ValueError(
                        "The following iteration(s) are not available: "
                        f"{missing_iter}. \n"
                        f"Available iteration(s): {available_iter}")

        # Plot weight distributions
        fig, ax = plt.subplots(1, 1, figsize=(4, 3))
        for iter_mlacs in mlacs_iter_arr:
            self._plot_weight_distribution(iter_mlacs, ax, normalize=False)
        ax.legend(frameon=False, loc='best')
        fig.tight_layout()

        dynamic_savename = f"{savename or 'plot_weights'}.pdf"
        fig.savefig(dynamic_savename, bbox_inches='tight')

        if show:
            os.system('xdg-open '+dynamic_savename)



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import shutil
import shlex
from pathlib import Path
from subprocess import run, PIPE

import numpy as np
from ase.io.lammpsdata import write_lammps_data
from ase.atoms import Atoms

from ..core.manager import Manager
from .descriptor import Descriptor, combine_reg
from ..utilities.io_lammps import (LammpsInput, LammpsBlockInput,
                                   get_lammps_command)


default_snap = {"twojmax": 8,
                "rfac0": 0.99363,
                "rmin0": 0.0,
                "switchflag": 1,
                "bzeroflag": 1,
                "wselfallflag": 0}

default_so3 = {"nmax": 4,
               "lmax": 4,
               "alpha": 1.0}


# ========================================================================== #
# ========================================================================== #
class MliapDescriptor(Descriptor):
    """
    Interface to the MLIAP potential of LAMMPS.

    Parameters
    ----------
    atoms : :class:`ase.atoms`
        Reference structure, with the elements for the descriptor

    rcut: :class:`float`
        The cutoff of the descriptor, in angstrom
        Default 5.0

    parameters: :class:`dict`
        A dictionnary of parameters for the descriptor input

        If the `style` is set to `snap`, then the default values are

        - twojmax = 8
        - rfac0 = 0.99363
        - rmin0 = 0.0
        - switchflag = 1
        - bzeroflag = 1
        - wselfallflag = 0

        If the `style` is set to `so3`, then the default values are

        - nmax = 4
        - lmax = 4
        - alpha = 1.0

    model: :class:`str`
        The type of model use. Can be either 'linear' or 'quadratic'
        Default `linear`

    style: :class:`str`
        The style of the descriptor used. Can be either 'snap' or 'so3'
        Default 'snap'

    alpha: :class:`float`
        The multiplication factor to the regularization parameter for
        ridge regression.
        Default 1.0

    Examples
    --------
    """
    def __init__(self, atoms, rcut=5.0, parameters={},
                 model="linear", style="snap", alpha=1.0,
                 prefix='MLIAP',
                 **kwargs):
        self.chemflag = parameters.pop("chemflag", False)
        Descriptor.__init__(self, atoms, rcut, alpha, prefix=prefix, **kwargs)

        self.model = model
        self.style = style

        # Initialize the parameters for the descriptors
        self.radelems = parameters.pop("radelems", None)
        if self.radelems is None:
            self.radelems = np.array([0.5 for i in self.elements])
        self.welems = parameters.pop("welems", None)
        if self.welems is None:
            self.welems = np.array(self.Z) / np.sum(self.Z)
        if self.style == "snap":
            self.params = default_snap
            self.params.update(parameters)
            if self.chemflag:
                self.params["bnormflag"] = 1
            twojmax = self.params["twojmax"]
            if twojmax % 2 == 0:
                m = 0.5 * twojmax + 1
                self.ndesc = int(m * (m+1) * (2*m+1) / 6)
            else:
                m = 0.5 * (twojmax + 1)
                self.ndesc = int(m * (m+1) * (m+2) / 3)
            if self.chemflag:
                self.ndesc *= self.nel**3
        elif self.style == "so3":
            self.params = default_so3
            self.params.update(parameters)
            nmax = self.params["nmax"]
            lmax = self.params["lmax"]
            self.welems /= self.welems.min()
            self.ndesc = int(nmax * (nmax + 1) * (lmax + 1) / 2)
        if self.model == "quadratic":
            self.ndesc += int(self.ndesc * (self.ndesc + 1) / 2)
        self.ncolumns = int(self.nel * (self.ndesc + 1))

        self.cmd = get_lammps_command()

# ========================================================================== #
    @Manager.exec_from_path
    def compute_descriptor(self, atoms, forces=True, stress=True):
        """
        """
        nat = len(atoms)

        lmp_atfname = "atoms.lmp"
        self._write_lammps_input(atoms.get_pbc())
        self._write_mlip_params()

        amat_e = np.zeros((1, self.ncolumns))
        amat_f = np.zeros((3 * nat, self.ncolumns))
        amat_s = np.zeros((6, self.ncolumns))
        write_lammps_data(lmp_atfname,
                          atoms,
                          specorder=self.elements.tolist())
        self._run_lammps(lmp_atfname)
        bispectrum = np.loadtxt("descriptor.out",
                                skiprows=4)

        bispectrum[-6:, 1:-1] /= -atoms.get_volume()
        amat_e[0] = bispectrum[0, 1:-1]
        amat_f = bispectrum[1:3*nat+1, 1:-1]
        amat_s = bispectrum[3*nat+1:, 1:-1]

        np.save("amat_e.npy", amat_e)
        np.save("amat_f.npy", amat_f)
        np.save("amat_s.npy", amat_s)

        self.cleanup()
        res = dict(desc_e=amat_e,
                   desc_f=amat_f,
                   desc_s=amat_s)
        return res

# ========================================================================== #
    @Manager.exec_from_path
    def compute_descriptors(self, atoms, forces=True, stress=True):
        """
        Compute the descriptors of multiples structures in one Lammps launch
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        if len(atoms) == 1:
            return super().compute_descriptors(atoms,
                                               forces=forces,
                                               stress=stress)
        if Path("Atoms").exists():
            shutil.rmtree("Atoms")
        Path("Atoms").mkdir()

        for i, at in enumerate(atoms):
            if np.any(at.get_pbc() != atoms[0].get_pbc()):
                raise ValueError("PBC cannot change between states")
            lmp_atfname = f"Atoms/atoms{i+1}.lmp"
            write_lammps_data(lmp_atfname,
                              at,
                              specorder=self.elements.tolist())
        self._write_mlip_params()
        self._write_looping_lammps_input(atoms)

        self._run_lammps(lmp_atfname)

        desc = []
        for i in range(len(atoms)):
            nat = len(atoms[i])
            amat_e = np.zeros((1, self.ncolumns))
            amat_f = np.zeros((3 * nat, self.ncolumns))
            amat_s = np.zeros((6, self.ncolumns))

            bispectrum = np.loadtxt(f"descriptor{i+1}.out",
                                    skiprows=4)
            bispectrum[-6:, 1:-1] /= -atoms[i].get_volume()

            amat_e[0] = bispectrum[0, 1:-1]
            amat_f = bispectrum[1:3*nat+1, 1:-1]
            amat_s = bispectrum[3*nat+1:, 1:-1]

            np.save("amat_e.npy", amat_e)
            np.save("amat_f.npy", amat_f)
            np.save("amat_s.npy", amat_s)

            desc.append(dict(desc_e=amat_e,
                             desc_f=amat_f,
                             desc_s=amat_s))
        self.loop_cleanup(len(atoms))
        return desc

# ========================================================================== #
    @Manager.exec_from_path
    def _write_looping_lammps_input(self, atoms):
        """
        Write one lammps file for all the config in atoms
        """
        # Write the input file
        txt = "LAMMPS input file for extracting MLIP descriptors"
        lmp_in = LammpsInput(txt)

        block = LammpsBlockInput("init", "Initialization")
        block("loop", f"variable a loop {len(atoms)}")
        block("label", "label loopstart")

        pbc_txt = "{0} {1} {2}".format(
                *tuple("sp"[int(x)] for x in atoms[0].get_pbc()))
        block("boundary", f"boundary {pbc_txt}")
        block("atom_style", "atom_style  atomic")
        block("units", "units metal")
        block("read_data", "read_data Atoms/atoms${a}.lmp")
        for i, m in enumerate(self.masses):
            block(f"mass{i}", f"mass   {i+1} {m}")
        lmp_in("init", block)

        block = LammpsBlockInput("interaction", "Interactions")
        block("pair_style", f"pair_style zero {2*self.rcut}")
        block("pair_coeff", "pair_coeff  * *")
        lmp_in("interaction", block)

        block = LammpsBlockInput("fake_dynamic", "Fake dynamic")
        block("thermo", "thermo 100")
        block("timestep", "timestep 0.005")
        block("neighbor", "neighbor 1.0 bin")
        block("neigh_modify", "neigh_modify once no every 1 delay 0 check yes")
        lmp_in("fake_dynamic", block)

        block = LammpsBlockInput("compute", "Compute")
        if self.style == "snap":
            style = "sna"
        elif self.style == "so3":
            style = "so3"
        txt = f"compute ml all mliap descriptor {style} " + \
              f"{self.prefix}.descriptor model {self.model}"
        block("compute", txt)
        block("fix", "fix ml all ave/time 1 1 1 c_ml[*] " +
              "file descriptor${a}.out mode vector")
        block("run", "run 0")
        lmp_in("compute", block)

        block = LammpsBlockInput("loopback", "Loop Back")
        block("clear data", "clear")
        block("iterate", "next a")
        block("loopend", "jump SELF loopstart")
        lmp_in("loopback", block)

        with open("lammps_input.in", "w") as fd:
            fd.write(str(lmp_in))

# ========================================================================== #
    @Manager.exec_from_path
    def _write_lammps_input(self, pbc):
        """
        """
        txt = "LAMMPS input file for extracting MLIP descriptors"
        lmp_in = LammpsInput(txt)

        block = LammpsBlockInput("init", "Initialization")
        block("clear", "clear")
        pbc_txt = "{0} {1} {2}".format(*tuple("sp"[int(x)] for x in pbc))
        block("boundary", f"boundary {pbc_txt}")
        block("atom_style", "atom_style  atomic")
        block("units", "units metal")
        block("read_data", "read_data atoms.lmp")
        for i, m in enumerate(self.masses):
            block(f"mass{i}", f"mass   {i+1} {m}")
        lmp_in("init", block)

        block = LammpsBlockInput("interaction", "Interactions")
        block("pair_style", f"pair_style zero {2*self.rcut}")
        block("pair_coeff", "pair_coeff  * *")
        lmp_in("interaction", block)

        block = LammpsBlockInput("fake_dynamic", "Fake dynamic")
        block("thermo", "thermo 100")
        block("timestep", "timestep 0.005")
        block("neighbor", "neighbor 1.0 bin")
        block("neigh_modify", "neigh_modify once no every 1 delay 0 check yes")
        lmp_in("fake_dynamic", block)

        block = LammpsBlockInput("compute", "Compute")
        if self.style == "snap":
            style = "sna"
        elif self.style == "so3":
            style = "so3"
        txt = f"compute ml all mliap descriptor {style} " + \
              f"{self.prefix}.descriptor model {self.model}"
        block("compute", txt)
        block("fix", "fix ml all ave/time 1 1 1 c_ml[*] " +
              "file descriptor.out mode vector")
        block("run", "run 0")
        lmp_in("compute", block)

        with open("lammps_input.in", "w") as fd:
            fd.write(str(lmp_in))

# ========================================================================== #
    @Manager.exec_from_path
    def _run_lammps(self, lmp_atoms_fname):
        '''
        Function that call LAMMPS to extract the descriptor and gradient values
        '''
        lmp_cmd = f"{self.cmd} -in lammps_input.in -log none -sc lmp.out"
        lmp_handle = run(shlex.split(lmp_cmd),
                         stderr=PIPE)

        # There is a bug in LAMMPS that makes compute_mliap crashes at the end
        if lmp_handle.returncode != 0:
            pass
            """
            msg = "LAMMPS stopped with the exit code \n" + \
                  f"{lmp_handle.stderr.decode()}"
            raise RuntimeError(msg)
            """

# ========================================================================== #
    @Manager.exec_from_path
    def cleanup(self):
        '''
        Function to cleanup the LAMMPS files used
        to extract the descriptor and gradient values
        '''
        Path("lmp.out").unlink(missing_ok=True)
        Path("descriptor.out").unlink(missing_ok=True)
        Path("lammps_input.in").unlink(missing_ok=True)
        Path("atoms.lmp").unlink(missing_ok=True)

# ========================================================================== #
    @Manager.exec_from_path
    def loop_cleanup(self, nconf):
        '''
        Function to cleanup the LAMMPS files used
        to extract the descriptor and gradient values
        '''
        Path("lmp.out").unlink(missing_ok=True)
        Path("lammps_input.in").unlink(missing_ok=True)
        shutil.rmtree("Atoms")
        for i in range(nconf):
            Path(f"descriptor{i+1}.out").unlink(missing_ok=True)

# ========================================================================== #
    @Manager.exec_from_path
    def _write_mlip_params(self):
        """
        Function to write the mliap.descriptor parameter files of the MLIP
        """
        with open(f"{self.prefix}.descriptor", "w") as f:
            f.write(self.get_mlip_params())

# ========================================================================== #
    def get_mlip_params(self):
        elements = self.elements

        s = ("# ")
        # Adding a commment line to know what elements are fitted here
        for el in elements:
            s += ("{:} ".format(el))
        s += ("MLIP parameters\n")
        s += (f"# Descriptor:  {self.style}\n")
        s += (f"# Model:       {self.model}\n")
        s += ("\n")
        s += (f"rcutfac         {self.rcut}\n")
        for key in self.params.keys():
            s += (f"{key:12}    {self.params[key]}\n")
        s += ("\n\n\n")
        s += (f"nelems      {len(elements)}\n")
        s += ("elems       ")
        for n in range(len(elements)):
            s += (elements[n] + " ")
        s += ("\n")
        s += ("radelems   ")
        for n in range(len(elements)):
            s += (f" {self.radelems[n]}")
        s += ("\n")
        s += ("welems    ")
        for n in range(len(elements)):
            s += (f"  {self.welems[n]}")
        s += ("\n")

        if self.style == "snap" and self.chemflag:
            s += ("\n\n")
            s += ("chemflag     1\n")
            s += ("bnormflag    1\n")
        return s

# ========================================================================== #
    @Manager.exec_from_path
    def write_mlip(self, coefficients):
        """
        """
        filepath = Path(self.get_filepath('.model'))
        if filepath.is_file():
            filepath.unlink()
        fname = filepath.relative_to(self.path)

        with open(filepath, "w") as fd:
            fd.write("# ")
            fd.write(" ".join(self.elements))
            fd.write(" MLIP parameters\n")
            fd.write(f"# Descriptor   {self.style}\n")
            fd.write("\n")

            fd.write("# nelems   ncoefs\n")
            fd.write(f"{self.nel} {self.ndesc + 1}\n")
            np.savetxt(fd, coefficients, fmt="%35.30f")
        return fname

# ========================================================================== #
    @Manager.exec_from_path
    def get_coef(self, filename=None):
        """
        Read MLIP coefficients from a file.
        """
        if filename:
            filename = Path(filename)
        else:
            filename = Path(self.get_filepath('.model'))

        if not filename.is_file():
            filename = filename.absolute()
            raise FileNotFoundError(f"File {filename} does not exist")

        with open(filename, "r") as fd:
            lines = fd.readlines()

        coefs = []
        for line in lines:
            line = line.strip()
            if line.startswith('#') or len(line) == 0:
                continue
            line = line.split()
            if len(line) == 2:  # Consistency check: nel, ndesc+1
                assert int(line[0]) == self.nel, "The descriptor changed"
                assert int(line[1]) == self.ndesc+1, "The descriptor changed"
                continue
            coefs.append(float(line[0]))
        return coefs

# ========================================================================== #
    def _regularization_matrix(self):
        # no regularization for the intercept
        d2 = [np.zeros((self.nel, self.nel))]
        d2.append(np.eye(self.ncolumns - self.nel))
        return combine_reg(d2)

# ========================================================================== #
    def get_pair_style(self):
        if self.style == "snap":
            style = "sna"
        elif self.style == "so3":
            style = "so3"
        modelfile = self.get_filepath('.model')
        descfile = self.subdir / f"{self.prefix}.descriptor"
        pair_style = f"mliap model {self.model} {modelfile} " + \
                     f"descriptor {style} {descfile}"
        return pair_style

# ========================================================================== #
    def get_pair_coeff(self):
        return [f"* * {' '.join(self.elements)}"]

# ========================================================================== #
    def get_pair_style_coeff(self):
        return self.get_pair_style(), self.get_pair_coeff()

# ========================================================================== #
    def __str__(self):
        txt = " ".join(self.elements)
        txt += f" {self.style} MLIAP descriptor,"
        txt += f" rcut = {self.rcut}"
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = f"{self.style} MLIAP descriptor\n"
        txt += f"{(len(txt) - 1) * '-'}\n"
        txt += "Elements :\n"
        txt += " ".join(self.elements) + "\n"
        txt += "Parameters :\n"
        txt += f"rcut                {self.rcut}\n"
        txt += f"chemflag            {self.chemflag}\n"
        for key, val in self.params.items():
            txt += f"{key:12}        {val}\n"
        txt += f"dimension           {self.ncolumns}\n"
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import sys
import os
import shutil
import warnings
import logging

from pathlib import Path
from subprocess import run, PIPE
import shlex

import numpy as np
from ase import Atoms
from ..utilities import make_dataframe

from ase.io import read
from ase.io.lammpsdata import write_lammps_data
from ase.data import atomic_masses, chemical_symbols

from ..core.manager import Manager
from .descriptor import Descriptor
from ..utilities.io_lammps import (LammpsInput, LammpsBlockInput,
                                   get_lammps_command)

try:
    import pandas as pd
    ispandas = True
except ImportError:
    ispandas = False

# Tensorflow Warning if not using GPU
warnings.filterwarnings("ignore", category=Warning, module="tensorflow")
try:
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # Remove GPU warning for tf
    import tensorflow as tf  # noqa
    tf.get_logger().setLevel(logging.ERROR)  # Remove a warning
    istf = True
except ImportError:
    istf = False

try:
    import pyace
    from pyace.generalfit import GeneralACEFit
    from pyace import create_multispecies_basis_config
    from pyace.metrics_aggregator import MetricsAggregator
    ispyace = True
except ImportError:
    ispyace = False

def_bconf = {'deltaSplineBins': 0.001,
             'embeddings': {"ALL": {'npot': 'FinnisSinclairShiftedScaled',
                                    'fs_parameters': [1, 1, 1, 0.5],
                                    'ndensity': 2}},
             'bonds': {"ALL": {'radbase': 'ChebExpCos',
                               'NameOfCutoffFunction': 'cos',
                               'radparameters': [5.25],
                               'rcut': 7,
                               'dcut': 0.01}},
             'functions': {"ALL": {'nradmax_by_orders': [15, 3, 2],
                                   'lmax_by_orders': [0, 2, 2]}}}

def_loss = {'kappa': 0.3, 'L1_coeffs': 1e-12, 'L2_coeffs': 1e-12,
            'w0_rad': 1e-12, 'w1_rad': 1e-12, 'w2_rad': 1e-12}

def_fitting = {'maxiter': 100, 'fit_cycles': 1, 'repulsion': 'auto',
               'optimizer': 'BFGS',
               'optimizer_options': {'disp': True, 'gtol': 0, 'xrtol': 0}}

def_backend = {'evaluator': 'tensorpot', 'parallel_mode': 'serial',
               'batch_size': 100, 'display_step': 50}

# ========================================================================== #
# ========================================================================== #
class AceDescriptor(Descriptor):
    """
    Interface to the ACE potential using python-ace. Pace potential in LAMMPS.
    Usually a better generalization than SNAP but harder to train.
    Can be used with state at different pressure.
    Note : You will probably want to use 10-20 states since the
           fitting is relatively slow

    Parameters
    ----------
    atoms : :class:`ase.atoms` or list
        Reference structure, with the elements for the descriptor

    free_at_e : :class:`dict`
        The energy of one atom for every element.
        The atom MUST be isolated in a big box (eV/at)
        e.g. : {'Cu': 12.1283, 'O': 3.237}

    rcut: :class:`float`
        The cutoff of the descriptor, in angstrom
        Default 5.0

    tol_e: :class:`float`
        The tolerance on energy between ACE and DFT (meV/at)
        Default `5`

    tol_f: :class:`float`
        The tolerance on forces between ACE and DFT (meV/ang)
        Default `25`

    bconf_dict: :class:`dict`
        A dictionnary of parameters for the BBasisConfiguration
        The default values are
            - deltaSplineBins: 0.001
            - elements: Found dynamically (e.g. Cu2O: ['Cu', 'O'])
            - embeddings: {"ALL": {npot: 'FinnisSinclairShiftedScaled',
                                   fs_parameters: [1, 1, 1, 0.5],
                                   ndensity: 2}}
            - bonds: {"ALL": {radbase: 'ChebExpCos',
                              NameOfCutoffFunction: 'cos',
                              radparameters: [5.25],
                              rcut: 7,
                              dcut: 0.01}}
            - functions: {"ALL": {nradmax_by_orders: [15, 3, 2],
                                  lmax_by_orders: [0, 2, 2]}}

    loss_dict: :class:`dict`
        A dictionnary of parameters for the loss function
        The default values are
            - kappa: 0.3
            - L1_coeffs: 1e-12
            - L2_coeffs: 1e-12
            - w0_rad: 1e-12
            - w1_rad: 1e-12
            - w2_rad: 1e-12

    fitting_dict: :class:`dict`
        A dictionnary of parameters for the minimization
        The default values are
            - weighting: MBAR|Uniform otherwise
            - loss: loss_dict
            - maxiter: 100
            - fit_cycles: 1
            - repulsion: 'auto'
            - optimizer: 'BFGS'
            - optimizer_options: {disp: True, gtol: 0, xrtol: 0}

    backend_dict: :class:`dict`
        A dictionnary of parameters for the backend
        The default values are
            - evaluator: 'tensorpot'
            - parallel_mode: 'parallel'
            - n_workers : None
            - batch_size: 100
            - display_step: 50
    """
    def __init__(self, atoms, free_at_e, rcut=5.0, tol_e=5, tol_f=25,
                 bconf_dict=None, loss_dict=None, fitting_dict=None,
                 backend_dict=None, nworkers=None):

        if not ispyace:
            raise ValueError("Could not import pyace")

        self.cmd = get_lammps_command()
        self._verify_dependency()

        Descriptor.__init__(self, atoms, rcut)

        self.prefix = "ACE"
        self.desc_name = "ACE"
        self.n_fit_attempt = 3
        self.db_fn = "ACE.pckl.gzip"
        self.pot_fn = "ACE.yace"
        self.tol_e = tol_e
        self.tol_f = tol_f
        self.free_at_e = free_at_e
        bconf = def_bconf if bconf_dict is None else bconf_dict
        self.loss = def_loss if loss_dict is None else loss_dict
        self.fitting = def_fitting if fitting_dict is None else fitting_dict
        self.backend = def_backend if backend_dict is None else backend_dict

        # Kappa such that loss(tol_e) = loss(tol_f)
        self.loss['kappa'] = (tol_e)**2 / ((tol_e)**2 + 3*(tol_f)**2)
        self.fitting['loss'] = self.loss
        self.data = None

        if 'nworkers' not in self.backend and nworkers is not None:
            self.backend['parallel_mode'] = "process"
            self.backend['nworkers'] = nworkers
        if 'elements' not in self.loss:
            bconf['elements'] = self.elements

        # Set bconf['rcut'] according to rcut if rcut is given
        if rcut is not None:
            for bond_type, bond_info in bconf['bonds'].items():
                bond_info['rcut'] = rcut
        else:  # Set rcut according to bconf['rcut'] if rcut is not given
            rcut = 0
            for bond_type, bond_info in bconf['bonds'].items():
                if bond_info['rcut'] > rcut:
                    rcut = bond_info['rcut']
        self.rcut = rcut

        self.bconf = create_multispecies_basis_config(bconf)
        self.acefit = None
        self.log = None

# ========================================================================== #
    @Manager.exec_from_subdir
    def redirect_logger(self):
        """
        Redirect python used by pyace to pyace.log
        """
        file_handler = logging.FileHandler(self.subdir / 'pyace.log')
        # A lot of loggers are involved with pyace
        loggers = [
            __name__,
            'pyace.generalfit',
            'pyace.preparedata',
            'pyace.fitadapter',
            'pyace.metrics_aggregator',
            'numexpr.utils',
            'tensorpotential.fit',
        ]
        logging.getLogger('numexpr.utils').setLevel(logging.CRITICAL)

        # Loop through the loggers and reset the file handler for each
        for logger_name in loggers:
            logger = logging.getLogger(logger_name)
            while logger.handlers:
                logger.removeHandler(logger.handlers[0])
            logger.addHandler(file_handler)
            logger.propagate = False
        self.log = logging.getLogger(__name__)

# ========================================================================== #
    def get_mlip_params(self):
        """
        """
        raise NotImplementedError

# ========================================================================== #
    @Manager.exec_from_path
    def get_mlip_file(self, folder):
        """
        Read MLIP coefficients from a file.
        """
        filename = Path(folder) / self.pot_fn

        if not filename.is_file():
            filename = filename.absolute()
            raise FileNotFoundError(f"File {filename} does not exist")
        return str(filename)

# ========================================================================== #
    def prepare_wf(self, wf, forces):
        """
        Reshape wf from a flat array to a list of np.array where each
        np.array correspond to a conf i and is of size self.natoms[i]
        """
        new_wf = []
        curr_index = 0
        for f in forces:
            new_wf.append(wf[curr_index:curr_index+len(f)*3])
            curr_index += len(f)*3
        new_wf = [np.sum(np.reshape(new_wf[i], [-1, 3]), axis=1)
                  for i in range(len(new_wf))]
        return new_wf

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def fit(self, atoms, weights, name=None):
        """
        """
        natoms = [len(at) for at in atoms]
        energy = [at.get_potential_energy() - self.calc_free_e(at)
                  for at in atoms]
        forces = [at.get_forces() for at in atoms]
        if name is None:
            name = [f"config{i}" for i in range(len(atoms))]

        # Data preparation
        nconfs = len(natoms)
        we = weights[:nconfs].tolist()
        wf = weights[nconfs:-(nconfs)*6]
        wf = self.prepare_wf(wf, forces)

        atomic_env = self.compute_descriptors(atoms)

        # Dataframe preparation
        df = self.get_df()
        df = make_dataframe(
             df=df, name=name, atoms=atoms, atomic_env=atomic_env,
             energy=energy, forces=forces, we=we, wf=wf)
        df.to_pickle(self.workdir / self.subdir / self.db_fn,
                     compression="gzip")

        # Do the fitting
        if self.acefit is None:
            self.create_acefit()
        else:
            # Dubious implementation of GeneralACEFit,
            # No function to update dataframe once created ??
            self.create_acefit()

        # Note that self.fitting is the same object as self.acefit.fit_config
        real_maxiter = self.fitting['maxiter']
        nattempt = 0
        retry = True
        while retry:
            retry, niter_done = self.actual_fit()
            nattempt += 1
            self.fitting['maxiter'] -= niter_done
            if nattempt > self.n_fit_attempt or self.fitting['maxiter'] < 1:
                retry = False
            if retry:
                self.acefit.fit_config['fit_cycles'] += 1

        self.fitting['maxiter'] = real_maxiter

        fn_yaml = "interim_potential_best_cycle.yaml"
        yace_cmd = f"pace_yaml2yace {fn_yaml} -o {self.pot_fn}"
        self.mlip_model = Path.cwd() / self.pot_fn
        run(shlex.split(yace_cmd), stdout=sys.stdout, stderr=sys.stdout)

        if not Path(self.pot_fn).exists():
            msg = "The ACE fitting wasn't successful\n"
            msg += "If interim_potential_best_cycle.yaml doesn't exist "
            msg += f"in {Path().cwd()} then the ACEfit went wrong.\n"
            msg += f"Else, try this command '{yace_cmd}' inside "
            msg += f"{Path().cwd()}"
            raise RuntimeError(msg)

        return self.pot_fn

# ========================================================================== #
    def actual_fit(self):
        try:
            self.acefit.fit()
        except StopIteration:  # For Scipy < 1.11
            pass
        fit_res = self.acefit.fit_backend.fitter.res_opt

        # Fit failed due to precision loss
        retry = fit_res.status == 2 and not fit_res.success
        niter_done = self.acefit.current_fit_iteration
        return retry, niter_done

# ========================================================================== #
    @Manager.exec_from_subdir
    def set_restart_coefficient(self, mlip_subfolder):
        """
        Restart the calculation with the coefficient from this folder.
        This is because we cannot get back the coefficients from yace
        """
        if str(mlip_subfolder).endswith(".yace"):
            path = Path(mlip_subfolder).with_suffix(".yaml")
            if not path.exists():
                new_fn = "interim_potential_best_cycle.yaml"
                path = Path(mlip_subfolder).parent / new_fn
            if not path.exists():
                raise ValueError("Cannot find yaml file to get coefficients")
        fn = str(path)
        self.bconf.load(fn)
        fc = int(pyace.BBasisConfiguration(fn).metadata["_fit_cycles"])
        self.fitting['fit_cycles'] = fc+1

# ========================================================================== #
    @Manager.exec_from_subdir
    def create_acefit(self):
        """
        Creates the ACEFit Object. We need at least 1 conf.
        """
        def check_conv(last_fit_metric_data):
            """
            Function called after every fitting iteration to check convergence
            """
            iter_num = last_fit_metric_data["iter_num"]
            if iter_num == 0:
                MetricsAggregator.print_detailed_metrics(
                    last_fit_metric_data, title='Initial state:')
            elif iter_num % self.backend['display_step'] == 0:
                MetricsAggregator.print_extended_metrics(
                    last_fit_metric_data, title='INIT STATS:')
            e = last_fit_metric_data["rmse_epa"] * 1e3
            f = last_fit_metric_data["rmse_f_comp"] * 1e3
            msg = f"Step: {last_fit_metric_data['iter_num']}   "
            msg += f"RMSE Energy: {e:.4f} (meV/at), "
            msg += f"RMSE Forces: {f:.4f} (meV/ang)"
            self.log.info(msg)
            if iter_num > 0 and self.tol_e >= e and self.tol_f >= f:
                s = "Convergence reached:\n"
                s += f"RMSE of energy: {e:.8f} < {self.tol_e} (meV/at)\n"
                s += f"RMSE of forces: {f:.8f} < {self.tol_f} (meV/ang)\n"
                self.log.info(s)
                raise StopIteration("My convergence reached")

        self.callback = lambda val: check_conv(val)
        self.data = dict(filename=str(self.db_fn))
        self.acefit = GeneralACEFit(potential_config=self.bconf,
                                    fit_config=self.fitting,
                                    data_config=self.data,
                                    backend_config=self.backend)
        self.acefit.fit_backend.fit_metrics_callback = self.callback

# ========================================================================== #
    def calc_free_e(self, atoms):
        """
        Calculate the contribution which isn't the binding energy
        for this particular configuration
        """
        e = 0
        for atom in atoms:
            e_tmp = self.free_at_e[atom.symbol]
            e += e_tmp
        return e

# ========================================================================== #
    @Manager.exec_from_path
    def predict(self, atoms, coef, folder):
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        if Path("Atoms").exists():
            shutil.rmtree("Atoms")
        Path("Atoms").mkdir()

        clear_loop = False  # If we use clear or delete_atoms in lammps input
        for i, at in enumerate(atoms):
            if np.any(at.get_pbc() != atoms[0].get_pbc()):
                raise ValueError("PBC cannot change between states")
            if np.any(at.cell.array != atoms[0].cell.array):
                clear_loop = True
            lmp_atfname = f"Atoms/atoms{i+1}.lmp"

            write_lammps_data(lmp_atfname,
                              at,
                              specorder=self.elements.tolist())

        self._write_lammps_input(atoms, coef, is_clear=clear_loop)

        self._run_lammps(lmp_atfname)
        e, f, s = self._read_lammps_output(atoms)
        self.cleanup(len(atoms))

        if len(e) == 1:
            return e[0], f[0], s[0]

        return e, f, s

# ========================================================================== #
    @Manager.exec_from_path
    def _read_lammps_output(self, atoms):
        with open("lammps.out", "r") as f:
            lines = f.readlines()

        toparse = []
        tofind = "Step          Temp          E_pair         Press"
        for idx, line in enumerate(lines):
            if tofind in line:
                # "WARNING:" may appear between Step and value
                while "WARNING" in lines[idx+1]:
                    idx += 1
                toparse.append(lines[idx+1].split())
        if len(toparse) != len(atoms):
            raise ValueError("lammps.out a thermo for each atoms")

        e, f, s = [], [], []
        for i in range(len(toparse)):
            thermo, at = toparse[i], atoms[i]
            e_correction = self.calc_free_e(at)  # To only get cohesion energy
            bar2GPa = 1e-4
            e.append((float(thermo[2]) + e_correction)/len(at))
            f.append(np.array(read(f'forces{i+1}.out',
                              format='lammps-dump-text').get_forces()))
            s.append(np.loadtxt(f'stress{i+1}.out', skiprows=4)[:, 1]*bar2GPa)
        return e, f, s

# ========================================================================== #
    @Manager.exec_from_path
    def cleanup(self, natoms):
        '''
        Function to cleanup the LAMMPS files used
        to extract the descriptor and gradient values
        '''
        for i in range(natoms):
            Path(f"forces{i+1}.out").unlink()
            Path(f"stress{i+1}.out").unlink()
        Path("lammps.out").unlink()
        Path("lammps.in").unlink()
        shutil.rmtree("Atoms")

# ========================================================================== #
    @Manager.exec_from_path
    def _write_lammps_input(self, atoms, coef, is_clear=True):
        """
        Create one lammps input to evaluate E,F,S of multiple atoms
        Note : We need to use clear if the cell changes between atoms
               Else, we can use delete_atoms which doesn't reload potential
        """
        masses = [atomic_masses[chemical_symbols.index(el)]
                  for el in self.elements]

        # Write the input file
        txt = "LAMMPS input file for extracting MLIP E, F, S"
        lmp_in = LammpsInput(txt)

        block_loopstart = LammpsBlockInput("loopstart", "Loop starts")
        block_loopstart("loop", f"variable a loop {len(atoms)}")
        block_loopstart("label", "label loopstart")
        if is_clear:
            # loopstart, init, atoms_data, pot, compute1, compute2, loopback
            block_loopstart("clear data", "clear")
            lmp_in("loop", block_loopstart)
            read_data = "read_data Atoms/atoms${a}.lmp"
        else:
            # init, atoms_data, pot, compute1, loopstart, compute2, loopback
            read_data = "read_data Atoms/atoms1.lmp"

        block = LammpsBlockInput("init", "Initialization")
        pbc_txt = "{0} {1} {2}".format(
                *tuple("sp"[int(x)] for x in atoms[0].get_pbc()))
        block("boundary", f"boundary {pbc_txt}")
        block("atom_style", "atom_style  atomic")
        block("units", "units metal")
        block("read_data", read_data)
        for i, m in enumerate(masses):
            block(f"mass{i}", f"mass   {i+1} {m}")
        lmp_in("init", block)

        block = LammpsBlockInput("interaction", "Interactions")
        block("pair_style", "pair_style pace")
        pc = f"pair_coeff * * {coef} {' '.join(self.elements)}"
        block("pair_coeff", pc)
        lmp_in("interaction", block)

        block = LammpsBlockInput("compute1", "Compute")
        block("thermo_style", "thermo_style custom step temp epair press")
        block("cp", "compute svirial all pressure NULL virial")
        lmp_in("compute1", block)

        if not is_clear:
            block_loopstart("delat", "delete_atoms group all")
            block_loopstart("rd", "read_data Atoms/atoms${a}.lmp add append")
            lmp_in("loop", block_loopstart)

        block = LammpsBlockInput("compute2", "Compute")
        block("fs", "fix svirial${a} all ave/time 1 1 1 c_svirial" +
              " file stress${a}.out mode vector")
        block("dump", "dump dump_force${a} all custom 1 forces${a}.out" +
              " id type x y z fx fy fz")
        block("run", "run 0")
        lmp_in("compute2", block)

        block = LammpsBlockInput("loopback", "Loop Back")
        block("iterate", "next a")
        block("loopend", "jump SELF loopstart")
        lmp_in("loopback", block)

        with open("lammps.in", "w") as fd:
            fd.write(str(lmp_in))

# ========================================================================== #
    def _verify_dependency(self):
        """
        """
        s = ""
        if not ispandas:
            s += "Pandas package error.\n"
            s += "Can you import mlacs THEN pandas.\n\n"

        if not istf:
            s += "Tensorflow package error.\n"
            s += "Please install Tensorflow to minimize using CPU.\n\n"

        try:
            import tensorpotential  # noqa
            gpus = tf.config.experimental.list_physical_devices('GPU')
            if len(gpus) == 0:
                # ON: Not great as MLACS logger hasn't been initialized
                mlacs_log = logging.getLogger('mlacs')
                mlacs_log.warn("No GPUs found, will fit on CPUs\n")
        except ImportError:
            s += "Tensorpotential package error.\n"
            s += "Please install Tensorpotential from:\n"
            s += "https://github.com/ICAMS/TensorPotential.\n\n"

        # Test that lammps-user-pace is installed by looking at lmp -help
        lmp_help = shlex.split(self.cmd + " -help")
        grep_pace = shlex.split(r"grep '\Wpace'")
        res = run(lmp_help, stdout=PIPE)
        out = run(grep_pace, input=res.stdout, stdout=PIPE)
        if out.returncode:  # There was no mention of pace in pair_style
            s += "Pace style not found.\n"
            s += "You can install Pace for Lammps from:\n"
            s += "https://github.com/ICAMS/lammps-user-pace\n\n"

        if s:  # If any import error
            raise ImportError(s)

# ========================================================================== #
    def compute_descriptor(self, atoms, forces=True, stress=False):
        """
        Get the Pyace.catomic.AtomicEnvironemnt
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]

        atomic_env = []
        pyacecalc = pyace.asecalc.PyACECalculator(self.bconf, fast_nl=True)
        for at in atoms:
            atomic_env.append(pyacecalc.get_atomic_env(at))
        return atomic_env

# ========================================================================== #
    @Manager.exec_from_path
    def _run_lammps(self, lmp_atoms_fname):
        '''
        Function that call LAMMPS to extract the descriptor and gradient values
        '''
        lammps_command = self.cmd + ' -in lammps.in -log none -sc lammps.out'
        lmp_handle = run(shlex.split(lammps_command),
                         stderr=PIPE)

        # There is a bug in LAMMPS that makes compute_mliap crashes at the end
        if lmp_handle.returncode != 0:
            msg = "LAMMPS stopped with the exit code \n" + \
                  f"{lmp_handle.stderr.decode()}"
            raise RuntimeError(msg)

# ========================================================================== #
    @Manager.exec_from_path
    def _write_mlip_params(self):
        """
        """
        raise NotImplementedError("No params for ACE... yet")

# ========================================================================== #
    def get_pair_style_coeff(self):
        """
        """
        pair_style = self.get_pair_style()
        pair_coeff = self.get_pair_coeff()
        return pair_style, pair_coeff

# ========================================================================== #
    def get_pair_style(self):
        """
        """
        pair_style = "pace"
        return pair_style

# ========================================================================== #
    def get_pair_coeff(self):
        """
        """
        acefile = self.get_filepath('.yace')
        pair_coeff = [f"* * {acefile} " +
                      ' '.join(self.elements)]
        return pair_coeff

# ========================================================================== #
    def _ace_opt_str(self):
        raise Warning("No options/params allowed for this version of ACE")
        return ""

# ========================================================================== #
    @Manager.exec_from_path
    def get_df(self):
        """
        Return the pandas.DataFrame object associated to this AceDescriptor
        """
        db_fn = Path(self.get_filepath(".pckl.gzip"))
        if db_fn.exists():
            df = pd.read_pickle(self.db_fn, compression="gzip")
        else:
            df = None
        return df

# ========================================================================== #
    def __str__(self):
        txt = " ".join(self.elements)
        txt += " ACE descriptor,"
        txt += f" rcut = {self.rcut}"
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = "ACE descriptor\n"
        txt += f"{(len(txt) - 1) * '-'}\n"
        txt += f"Rcut: {self.rcut} ang\n"
        txt += f"Free atom energy (eV/at): {self.free_at_e}\n"
        txt += f"Dataframe : {self.db_fn}\n"
        txt += f"Tolerance on e : {self.tol_e} (meV/at)\n"
        txt += f"Tolerance on f : {self.tol_f} (meV/ang)\n"
        txt += f"Fitting: {self.fitting}\n"
        txt += f"Backend: {self.backend}\n"
        txt += f"BBasisConfiguration: {self.bconf}\n"
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import sys
import shutil
import pytest
import subprocess
from pathlib import Path


def mlacs_examples():
    root = Path().absolute().parents[0] / 'examples'
    expls = [f.name for f in root.iterdir() if f.name.startswith('mlacs_')]
    not_tested_expl = ['Abinit', 'QEspresso', '32Ag_EMT_300K10GPa_Snap',
                       '500Cu_EMT_400K50GPax5_SnapMBAR']
    for expl in expls:
        if any(_ in expl for _ in not_tested_expl):
            expls.remove(expl)
    return expls


@pytest.mark.examples
@pytest.mark.parametrize("example", mlacs_examples())
def test_mlacs_examples(example):
    root = Path().absolute().parents[0]
    file = root / 'examples' / example
    exe = sys.executable
    returncode = subprocess.call(f'{exe} {file}', shell=True)
    assert returncode == 0, \
        f'The example {example} is broken, please check it.'
    assert (root / 'tests' / f'{example.replace(".py", "")}').exists()
    shutil.rmtree(root / 'tests' / f'{example.replace(".py", "")}')


def post_examples():
    root = Path().absolute().parents[0] / 'examples'
    expls = [f.name for f in root.iterdir() if f.name.startswith('post_')]
    return expls


@pytest.mark.examples
@pytest.mark.parametrize("example", post_examples())
def test_mlacs_post_examples(example):
    prefix = example.replace('.py', '').replace('post_', '')
    root = Path().absolute().parents[0]
    file = root / 'examples' / f'mlacs_{prefix}.py'
    exe = sys.executable
    returncode = subprocess.call(f'{exe} {file}', shell=True)
    assert returncode == 0, \
        f'The example mlacs_{prefix}.py is broken, please check it.'
    assert (root / 'tests' / f'mlacs_{prefix}').exists()
    file = root / 'examples' / example
    returncode = subprocess.call(f'{exe} {file}', shell=True)
    assert returncode == 0, \
        f'The example {example} is broken, please check it.'
    assert (root / 'tests' / f'mlacs_{prefix}' / f'{prefix}_plot.pdf').exists()
    shutil.rmtree(root / 'tests' / f'mlacs_{prefix}')


def ti_examples():
    root = Path().absolute().parents[0] / 'examples'
    expls = [f.name for f in root.iterdir() if f.name.startswith('neti_')
             or f.name.startswith('rsti_')]
    return expls


@pytest.mark.examples
@pytest.mark.parametrize("example", ti_examples())
def test_ti_examples(example):
    root = Path().absolute().parents[0]
    file = root / 'examples' / example
    exe = sys.executable
    returncode = subprocess.call(f'{exe} {file}', shell=True)
    assert returncode == 0, \
        f'The example {example} is broken, please check it.'
    assert (root / 'tests' / f'{example.replace(".py", "")}').exists()
    shutil.rmtree(root / 'tests' / f'{example.replace(".py", "")}')



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.units import GPa
from ase import Atoms

from . import MlipManager
from ..utilities import compute_correlation, create_link

default_parameters = {"method": "ols",
                      "lambda_ridge": 1e-8,
                      "hyperparameters": {},
                      "gridcv": {}}


# ========================================================================== #
# ========================================================================== #
class LinearPotential(MlipManager):
    """
    Potential that assume a linear relation between the descriptor and the
    energy.

    Parameters
    ----------
    descriptor: :class:`Descriptor`
        The descriptor used in the model.

    parameters: :class:`dict`
        The parameters for the fit.
        By default, the fit is a simple ordinary least squares.
        Ridge regression can be use by setting a dictionnary as
        ``dict(method=ridge, lambda_ridge=alpha)``, with alpha the ridge
        coefficient.

    weight: :class:`WeightingPolicy`
        Weight used for the fitting and calculation of properties.
        Default :class:`None`, which results in the use of uniform weights.

    Examples
    --------

    >>> from ase.io import read
    >>> confs = read('Trajectory.traj', index=':')
    >>>
    >>> from mlacs.mlip import SnapDescriptor, LinearPotential
    >>> desc = SnapDescriptor(confs[0], rcut=6.2, parameters=dict(twojmax=6))
    >>> mlip = LinearPotential(desc)
    >>>
    >>> mlip.update_matrices(confs)
    >>> mlip.train_mlip()
    """

    def __init__(self,
                 descriptor,
                 parameters={},
                 weight=None,
                 **kwargs):

        MlipManager.__init__(self,
                             descriptor,
                             weight,
                             **kwargs)

        self.parameters = default_parameters
        self.parameters.update(parameters)

        self.coefficients = None

        if self.parameters["method"] != "ols":
            if self.parameters["hyperparameters"] is None:
                hyperparam = {}
            else:
                hyperparam = self.parameters["hyperparameters"]
            hyperparam["fit_intercept"] = False
            self.parameters["hyperparameters"] = hyperparam
        self.can_use_weight = True

# ========================================================================== #
    def train_mlip(self):
        """
        Compute the coefficients of the MLIP, then write MLIP.

        Notes
        -----

        Local variables:
            - `amat`: ndarray of shape ((7+3*Nat)*N, K)
                Feature matrix
            - `ymat`: ndarray of shape ((7+3*Nat)*N,)
                Label vector
            - `W`: ndarray of shape ((7+3*Nat)*N,)
                Weighting matrix

        Where:
            - `K` is the number of descriptor components
            - `N` is the number of configurations
            - `Nat` is the number of atoms in each cell

        var = `amat`, `ymat`, `W` have a stacked structure
            +-------+
            |   e   |  <-- energy block
            +-------+
            |   f   |  <-- forces block
            +-------+
            |   s   |  <-- stresses block
            +-------+
        where each block has the shape of var_i, with i=e,f,s
        """
        self.weight.workdir = self.workdir
        self.weight.folder = self.folder
        self.weight.subfolder = self.subfolder
        self.descriptor.workdir = self.workdir
        self.descriptor.folder = self.folder
        self.descriptor.subfolder = self.subfolder

        msg = ''
        amat_e = self.amat_e / self.natoms[:, None]
        amat_f = self.amat_f
        amat_s = self.amat_s
        ymat_e = np.copy(self.ymat_e) / self.natoms
        ymat_f = self.ymat_f
        ymat_s = self.ymat_s

        # Division by amat.std : If de=1e2 and ds=1e5, we dont want to fit
        # 1000x more on the stress than on the energy. Careful ymat/AMAT.std
        amat = np.r_[amat_e / amat_e.std(),
                     amat_f / amat_f.std(),
                     amat_s / amat_s.std()]
        ymat = np.r_[ymat_e / amat_e.std(),
                     ymat_f / amat_f.std(),
                     ymat_s / amat_s.std()]

        W = self.weight.get_weights()
        amat = amat * W[:, np.newaxis]
        ymat = ymat * W
        if self.parameters["method"] == "ols":
            self.coefficients = np.linalg.lstsq(amat,
                                                ymat,
                                                None)[0]
        elif self.parameters["method"] == "ridge":
            lamb = self.parameters["lambda_ridge"]
            gamma = self.descriptor._regularization_matrix()
            ymat = amat.T @ ymat
            amat = amat.T @ amat + gamma * lamb
            self.coefficients = np.linalg.lstsq(amat,
                                                ymat,
                                                None)[0]

        else:
            msg = f"Fitting method {self.parameters['method']} " + \
                  "unknown"
            raise ValueError(msg)

        msg += "\nNumber of configurations for training: " + \
               f"{len(self.natoms):}\n"
        msg += "Number of atomic environments for training: " + \
               f"{self.natoms.sum():}\n\n"

        tmp_msg, weight_fn = self.weight.compute_weight(self.coefficients,
                                                        self.predict)

        msg += tmp_msg
        msg += self.compute_tests(amat_e, amat_f, amat_s,
                                  ymat_e, ymat_f, ymat_s)

        mlip_fn = self.descriptor.write_mlip(self.coefficients)
        create_link(self.subsubdir / weight_fn, self.subdir / weight_fn)
        create_link(self.subsubdir / mlip_fn, self.subdir / mlip_fn)

        if self.log:
            self.log.write(msg)

# ========================================================================== #
    def compute_tests(self, amat_e, amat_f, amat_s,
                      ymat_e, ymat_f, ymat_s):
        """
        Computed the weighted RMSE and MAE.
        """
        e_mlip = np.einsum('ij,j->i', amat_e, self.coefficients)
        f_mlip = np.einsum('ij,j->i', amat_f, self.coefficients)
        s_mlip = np.einsum('ij,j->i', amat_s, self.coefficients)

        w = None
        if len(self.weight.weight) > 0:
            w = self.weight.weight

        wf = np.array([])
        for i in range(len(w)):
            wf = np.append(wf, np.ones(self.natoms[i]*3)*(w[i]/3))

        res_E = compute_correlation(np.c_[ymat_e, e_mlip], weight=w)
        res_F = compute_correlation(np.c_[ymat_f, f_mlip], weight=wf)
        res_S = compute_correlation(np.c_[ymat_s, s_mlip]/GPa, weight=w)
        self.fit_res = np.c_[res_E, res_F, res_S]

        # Information to MLIP-Energy_comparison.dat
        header = f"Weighted rmse: {self.fit_res[0, 0]:.6f} eV/at,    " + \
                 f"Weighted mae: {self.fit_res[1, 0]:.6f} eV/at\n" + \
                 " True Energy           Predicted Energy"
        np.savetxt("MLIP-Energy_comparison.dat",
                   np.c_[ymat_e, e_mlip],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"Weighted rmse: {self.fit_res[0, 1]:.6f} eV/angs   " + \
                 f"Weighted mae: {self.fit_res[1, 1]:.6f} eV/angs\n" + \
                 " True Forces           Predicted Forces"

        np.savetxt("MLIP-Forces_comparison.dat",
                   np.c_[ymat_f, f_mlip],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"Weighted rmse: {self.fit_res[0, 2]:.6f} GPa       " + \
                 f"Weighted mae: {self.fit_res[1, 2]:.6f} GPa\n" + \
                 " True Stress           Predicted Stress"
        np.savetxt("MLIP-Stress_comparison.dat",
                   np.c_[ymat_s, s_mlip] / GPa,
                   header=header, fmt="%25.20f  %25.20f")

        # Message to Mlacs.log
        msg = f"Weighted RMSE Energy    {self.fit_res[0, 0]:.4f} eV/at\n"
        msg += f"Weighted MAE Energy     {self.fit_res[1, 0]:.4f} eV/at\n"
        msg += f"Weighted RMSE Forces    {self.fit_res[0, 1]:.4f} eV/angs\n"
        msg += f"Weighted MAE Forces     {self.fit_res[1, 1]:.4f} eV/angs\n"
        msg += f"Weighted RMSE Stress    {self.fit_res[0, 2]:.4f} GPa\n"
        msg += f"Weighted MAE Stress     {self.fit_res[1, 2]:.4f} GPa\n"
        return msg

# ========================================================================== #
    def get_calculator(self):
        """
        Initialize a ASE calculator from the model
        """
        from .calculator import MlipCalculator
        calc = MlipCalculator(self)
        return calc

# ========================================================================== #
    def predict(self, atoms, coef=None):
        """
        Predict energy (eV), forces (eV/ang) and stress (eV/ang**3) given
        desc which can be of type ase.Atoms or list of ase.Atoms.
        Can choose the coefficients to calculate with, or use the latest one
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        if coef is None:
            coef = self.coefficients
        assert coef is not None, 'The model has not been trained'

        if not isinstance(atoms[0], Atoms):
            raise NotImplementedError

        desc = self.descriptor.compute_descriptors(atoms)

        amat_e = [d['desc_e'] for d in desc]
        amat_f = [d['desc_f'] for d in desc]
        amat_s = [d['desc_s'] for d in desc]

        # We use the latest value coefficients to get the properties
        energy = np.einsum('nij,j->n',  amat_e, coef)
        forces = [np.einsum('ij,j->i', conf_f, coef) for conf_f in amat_f]
        stress = np.einsum('nij,j->ni', amat_s, coef)

        if len(energy) == 1:
            return energy[0], forces[0], stress[0]
        return energy, forces, stress

# ========================================================================== #
    def set_coefficients(self, coefficients):
        """
        """
        if coefficients is not None:
            assert len(coefficients) == self.descriptor.ncolumns
        self.coefficients = coefficients

# ========================================================================== #
    def __str__(self):
        txt = f"Linear potential with {str(self.descriptor)}"
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = "Linear potential\n"
        txt += "Parameters:\n"
        txt += "-----------\n"
        txt += f"Fit method :            {self.parameters['method']}\n"
        txt += "\n"
        txt += "Descriptor used in the potential:\n"
        txt += repr(self.descriptor)
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path

from ..utilities.miscellanous import execute_from


class Manager:
    """
    Base class for managing computation files and logs,

    Parameters
    ----------

    workdir: :class:`str`
        The working directory.
        It is fixed, and may be shared with other managers writing in it.

    folder: :class:`str`
        A folder inside the working directory, specific to this object.
        Its name is relative to workdir, and its absolute path is accessed
        with the subdir property.

    subfolder: :class:`str`
        A subfolder inside the folder directory.
        Its name is relative to subdir, and its absolute path is accessed
        with the subsubdir property.
        It can be dynamic, e.g. if we perform multiple calculations.

    prefix: :class:`str`
        A prefix for making file names.


    Examples
    --------

    workdir/prefix.traj
    ├─ folder1/
    │  ├─ subfolder1/prefix.traj
    ├─ folder2/
    │  ├─ subfolder1/
    │  ├─ subfolder2/

    path -> /workdir/folder1/subfolder2/
    filepath -> /workdir/folder1/subfolder2/prefix.traj

    subdir -> /workdir/folder1/
    subsubdir -> /workdir/folder1/subfolder2/
    """

    _workdir = Path('')
    _folder = ''
    _subfolder = ''
    _prefix = ''

    def __init__(self,
                 workdir='',
                 folder='',
                 subfolder='',
                 prefix='',
                 log=None,
                 **kwargs):

        self.workdir = workdir
        self.folder = folder
        self.subfolder = subfolder
        self.prefix = prefix
        self.log = log

    @property
    def workdir(self):
        """Starting directory for execution."""
        return self._workdir

    @workdir.setter
    def workdir(self, x):
        self._workdir = Path(x or '').absolute()

    @property
    def subdir(self):
        """The path of the sub-directory."""
        return self.workdir / self.folder

    @subdir.setter
    def subdir(self, x):
        self.folder = x

    @property
    def subsubdir(self):
        """The path of the sub-sub-directory."""
        return self.subdir / self.subfolder

    @subsubdir.setter
    def subsubdir(self, x):
        self.subfolder = x

    @property
    def folder(self):
        """Name of sub-directory, relative to workdir."""
        return self._folder

    @folder.setter
    def folder(self, x):
        x = x or ''
        if isinstance(x, Path):
            if x.is_absolute():
                if x == self.subdir:
                    return
                else:
                    self._folder = str(x.relative_to(self.workdir))
            else:
                self._folder = str(x)
        else:
            self._folder = x

    @property
    def subfolder(self):
        """Name of sub-sub-directory, relative to sub-directory."""
        return self._subfolder

    @subfolder.setter
    def subfolder(self, x):
        x = x or ''
        if isinstance(x, Path):
            if x.is_absolute():
                if x == self.subsubdir:
                    return
                else:
                    self._subfolder = str(x.relative_to(self.subdir))
            else:
                self._subfolder = str(x)
        else:
            self._subfolder = x

    @staticmethod
    def exec_from_workdir(func):
        """Decorator that executes a function from the working directory."""
        def wrapper(self, *args, **kwargs):
            self.workdir.mkdir(exist_ok=True, parents=True)
            with execute_from(self.workdir):
                result = func(self, *args, **kwargs)
            return result
        return wrapper

    @staticmethod
    def exec_from_subdir(func):
        """Decorator that executes a function from the subdirectory."""
        def wrapper(self, *args, **kwargs):
            self.subdir.mkdir(exist_ok=True, parents=True)
            with execute_from(self.subdir):
                result = func(self, *args, **kwargs)
            return result
        return wrapper

    @staticmethod
    def exec_from_subsubdir(func):
        """Decorator that executes a function from the subdirectory."""
        def wrapper(self, *args, **kwargs):
            self.subsubdir.mkdir(exist_ok=True, parents=True)
            with execute_from(self.subsubdir):
                result = func(self, *args, **kwargs)
            return result
        return wrapper

    # Make an alias
    path = subsubdir
    exec_from_path = exec_from_subsubdir

    @property
    def prefix(self):
        """Prefix for naming various files."""
        return self._prefix

    @prefix.setter
    def prefix(self, x):
        self._prefix = str(x or '')

    @property
    def filepath(self):
        """
        Absolute path of the sub-sub-directory,
        or path of a file named self.prefix.
        """
        if self.prefix:
            return self.subsubdir / self.prefix
        return str(self.subsubdir / 'X')[:-1]

    def get_filepath(self, suffix: str):
        """Construct a file name and return absolute path."""
        if self.prefix:
            return str(self.subsubdir / (self.prefix + suffix))
        return str(self.subsubdir / suffix)

    def iter_subsubdir(self, nmax=None, rootname=''):
        pass

    def iter_prefix(self, prefix=None, nmax=None, rootname=''):
        pass
